# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import Any, List, Optional, Tuple, Type, Union

from .._utils import create_actor_ref, to_binary
from ..aio import AioFileObject
from ..api import Actor
from ..context import BaseActorContext
from ..core import ActorRef, BufferRef, FileObjectRef, create_local_actor_ref
from ..debug import debug_async_timeout, detect_cycle_send
from ..errors import CannotCancelTask
from ..utils import dataslots, fix_all_zero_ip
from .allocate_strategy import AddressSpecified, AllocateStrategy
from .communication import Client, DummyClient, UCXClient
from .core import ActorCaller
from .message import (
    DEFAULT_PROTOCOL,
    ActorRefMessage,
    CancelMessage,
    ControlMessage,
    ControlMessageType,
    CopyToBuffersMessage,
    CopyToFileObjectsMessage,
    CreateActorMessage,
    DestroyActorMessage,
    ErrorMessage,
    HasActorMessage,
    ResultMessage,
    SendMessage,
    _MessageBase,
    new_message_id,
)
from .router import Router

DEFAULT_TRANSFER_BLOCK_SIZE = 4 * 1024**2


@dataslots
@dataclass
class ProfilingContext:
    task_id: str


class IndigenActorContext(BaseActorContext):
    __slots__ = ("_caller", "_lock")

    support_allocate_strategy = True

    def __init__(self, address: str | None = None):
        BaseActorContext.__init__(self, address)
        self._caller = ActorCaller()
        self._lock = asyncio.Lock()

    def __del__(self):
        self._caller.cancel_tasks()

    async def _call(
        self,
        address: str,
        message: _MessageBase,
        wait: bool = True,
        proxy_addresses: list[str] | None = None,
    ) -> Union[ResultMessage, ErrorMessage, asyncio.Future]:
        return await self._caller.call(
            Router.get_instance_or_empty(),
            address,
            message,
            wait=wait,
            proxy_addresses=proxy_addresses,
        )

    async def _call_with_client(
        self, client: Client, message: _MessageBase, wait: bool = True
    ) -> Union[ResultMessage, ErrorMessage, asyncio.Future]:
        # NOTE: used by copyto, cannot support proxy
        return await self._caller.call_with_client(client, message, wait)

    async def _call_send_buffers(
        self,
        client: UCXClient,
        local_buffers: list,
        meta_message: _MessageBase,
        wait: bool = True,
    ) -> Union[ResultMessage, ErrorMessage, asyncio.Future]:
        return await self._caller.call_send_buffers(
            client, local_buffers, meta_message, wait
        )

    @staticmethod
    def _process_result_message(message: Union[ResultMessage, ErrorMessage]):
        if isinstance(message, ResultMessage):
            return message.result
        else:
            raise message.as_instanceof_cause()

    async def _wait(self, future: asyncio.Future, address: str, message: _MessageBase):
        try:
            await asyncio.shield(future)
        except asyncio.CancelledError:
            try:
                await self.cancel(address, message.message_id)
            except CannotCancelTask:
                # cancel failed, already finished
                raise asyncio.CancelledError
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except
            pass
        return await future

    async def create_actor(
        self,
        actor_cls: Type[Actor],
        *args,
        uid=None,
        address: str | None = None,
        **kwargs,
    ) -> ActorRef:
        router = Router.get_instance_or_empty()
        address = address or self._address or router.external_address
        allocate_strategy = kwargs.get("allocate_strategy", None)
        if isinstance(allocate_strategy, AllocateStrategy):
            allocate_strategy = kwargs.pop("allocate_strategy")
        else:
            allocate_strategy = AddressSpecified(address)
        create_actor_message = CreateActorMessage(
            new_message_id(),
            actor_cls,
            to_binary(uid),
            args,
            kwargs,
            allocate_strategy,
            protocol=DEFAULT_PROTOCOL,
        )
        future = await self._call(address, create_actor_message, wait=False)
        result = await self._wait(future, address, create_actor_message)  # type: ignore
        return self._process_result_message(result)

    async def has_actor(self, actor_ref: ActorRef) -> bool:
        message = HasActorMessage(
            new_message_id(), actor_ref, protocol=DEFAULT_PROTOCOL
        )
        future = await self._call(
            actor_ref.address,
            message,
            wait=False,
            proxy_addresses=actor_ref.proxy_addresses,
        )
        result = await self._wait(future, actor_ref.address, message)  # type: ignore
        return self._process_result_message(result)

    async def destroy_actor(self, actor_ref: ActorRef):
        message = DestroyActorMessage(
            new_message_id(), actor_ref, protocol=DEFAULT_PROTOCOL
        )
        future = await self._call(
            actor_ref.address,
            message,
            wait=False,
            proxy_addresses=actor_ref.proxy_addresses,
        )
        result = await self._wait(future, actor_ref.address, message)  # type: ignore
        return self._process_result_message(result)

    async def kill_actor(self, actor_ref: ActorRef, force: bool = True):
        # get main_pool_address
        control_message = ControlMessage(
            new_message_id(),
            actor_ref.address,
            ControlMessageType.get_config,
            "main_pool_address",
            protocol=DEFAULT_PROTOCOL,
        )
        main_address = self._process_result_message(
            await self._call(actor_ref.address, control_message, proxy_addresses=actor_ref.proxy_addresses)  # type: ignore
        )
        real_actor_ref = await self.actor_ref(actor_ref)
        if real_actor_ref.address == main_address:
            raise ValueError("Cannot kill actor on main pool")
        stop_message = ControlMessage(
            new_message_id(),
            real_actor_ref.address,
            ControlMessageType.stop,
            # default timeout (3 secs) and force
            (3.0, force),
            protocol=DEFAULT_PROTOCOL,
        )
        # stop server
        result = await self._call(
            main_address, stop_message, proxy_addresses=actor_ref.proxy_addresses
        )
        return self._process_result_message(result)  # type: ignore

    async def actor_ref(self, *args, **kwargs):
        actor_ref = create_actor_ref(*args, **kwargs)
        connect_addr = actor_ref.address
        local_actor_ref = create_local_actor_ref(actor_ref.address, actor_ref.uid)
        if local_actor_ref is not None:
            return local_actor_ref
        message = ActorRefMessage(
            new_message_id(), actor_ref, protocol=DEFAULT_PROTOCOL
        )
        future = await self._call(
            actor_ref.address,
            message,
            wait=False,
            proxy_addresses=actor_ref.proxy_addresses,
        )
        result = await self._wait(future, actor_ref.address, message)
        res = self._process_result_message(result)
        if res.address != connect_addr:
            res.address = fix_all_zero_ip(res.address, connect_addr)
        return res

    async def send(
        self,
        actor_ref: ActorRef,
        message: Tuple,
        wait_response: bool = True,
        profiling_context: ProfilingContext | None = None,
    ):
        send_message = SendMessage(
            new_message_id(),
            actor_ref,
            message,
            protocol=DEFAULT_PROTOCOL,
            profiling_context=profiling_context,
        )

        # use `%.500` to avoid print too long messages
        with debug_async_timeout(
            "actor_call_timeout",
            "Calling %.500r on %s at %s timed out",
            send_message.content,
            actor_ref.uid,
            actor_ref.address,
        ):
            detect_cycle_send(send_message, wait_response)
            future = await self._call(
                actor_ref.address,
                send_message,
                wait=False,
                proxy_addresses=actor_ref.proxy_addresses,
            )
            if wait_response:
                result = await self._wait(future, actor_ref.address, send_message)  # type: ignore
                return self._process_result_message(result)
            else:
                return future

    async def cancel(self, address: str, cancel_message_id: bytes):
        message = CancelMessage(
            new_message_id(), address, cancel_message_id, protocol=DEFAULT_PROTOCOL
        )
        result = await self._call(address, message)
        return self._process_result_message(result)  # type: ignore

    async def wait_actor_pool_recovered(
        self, address: str, main_address: str | None = None
    ):
        if main_address is None:
            # get main_pool_address
            control_message = ControlMessage(
                new_message_id(),
                address,
                ControlMessageType.get_config,
                "main_pool_address",
                protocol=DEFAULT_PROTOCOL,
            )
            main_address = self._process_result_message(
                await self._call(address, control_message)  # type: ignore
            )

        # if address is main pool, it is never recovered
        if address == main_address:
            return

        control_message = ControlMessage(
            new_message_id(),
            address,
            ControlMessageType.wait_pool_recovered,
            None,
            protocol=DEFAULT_PROTOCOL,
        )
        self._process_result_message(await self._call(main_address, control_message))  # type: ignore

    async def get_pool_config(self, address: str):
        control_message = ControlMessage(
            new_message_id(),
            address,
            ControlMessageType.get_config,
            None,
            protocol=DEFAULT_PROTOCOL,
        )
        return self._process_result_message(await self._call(address, control_message))  # type: ignore

    @staticmethod
    def _gen_switch_to_copy_to_control_message(content: Any):
        return ControlMessage(
            message_id=new_message_id(),
            control_message_type=ControlMessageType.switch_to_copy_to,
            content=content,
        )

    @staticmethod
    def _gen_copy_to_buffers_message(content: Any):
        return CopyToBuffersMessage(message_id=new_message_id(), content=content)  # type: ignore

    @staticmethod
    def _gen_copy_to_fileobjs_message(content: Any):
        return CopyToFileObjectsMessage(message_id=new_message_id(), content=content)  # type: ignore

    async def _get_copy_to_client(self, router, address) -> Client:
        client = await self._caller.get_client(router, address)
        if isinstance(client, DummyClient) or hasattr(client, "send_buffers"):
            return client
        client_types = router.get_all_client_types(address)
        # For inter-process communication, the ``self._caller.get_client`` interface would not look for UCX Client,
        # we still try to find UCXClient for this case.
        try:
            client_type = next(
                client_type
                for client_type in client_types
                if hasattr(client_type, "send_buffers")
            )
        except StopIteration:
            return client
        else:
            return await self._caller.get_client_via_type(router, address, client_type)

    async def _get_client(self, address: str) -> Client:
        router = Router.get_instance()
        assert router is not None, "`copy_to` can only be used inside pools"
        if router.get_proxy(address):
            raise RuntimeError("Cannot run `copy_to` when enabling proxy")
        return await self._get_copy_to_client(router, address)

    async def copy_to_buffers(
        self,
        local_buffers: list,
        remote_buffer_refs: List[BufferRef],
        block_size: Optional[int] = None,
    ):
        address = remote_buffer_refs[0].address
        client = await self._get_client(address)
        if isinstance(client, UCXClient):
            message = [(buf.address, buf.uid) for buf in remote_buffer_refs]
            await self._call_send_buffers(
                client,
                local_buffers,
                self._gen_switch_to_copy_to_control_message(message),
            )
        else:
            # ``local_buffers`` will be divided into buffers of the specified block size for transmission.
            # Smaller buffers will be accumulated and sent together,
            # while larger buffers will be divided and sent.
            current_buf_size = 0
            one_block_data = []
            block_size = block_size or DEFAULT_TRANSFER_BLOCK_SIZE
            for i, (l_buf, r_buf) in enumerate(zip(local_buffers, remote_buffer_refs)):
                if current_buf_size + len(l_buf) < block_size:
                    one_block_data.append(
                        (r_buf.address, r_buf.uid, 0, len(l_buf), l_buf)
                    )
                    current_buf_size += len(l_buf)
                    continue
                last_start = 0
                while current_buf_size + len(l_buf) > block_size:
                    remain = block_size - current_buf_size
                    one_block_data.append(
                        (r_buf.address, r_buf.uid, last_start, remain, l_buf[:remain])
                    )
                    await self._call_with_client(
                        client, self._gen_copy_to_buffers_message(one_block_data)
                    )
                    one_block_data = []
                    current_buf_size = 0
                    last_start += remain
                    l_buf = l_buf[remain:]

                if len(l_buf) > 0:
                    one_block_data.append(
                        (r_buf.address, r_buf.uid, last_start, len(l_buf), l_buf)
                    )
                    current_buf_size = len(l_buf)

            if one_block_data:
                await self._call_with_client(
                    client, self._gen_copy_to_buffers_message(one_block_data)
                )

    async def copy_to_fileobjs(
        self,
        local_fileobjs: List[AioFileObject],
        remote_fileobj_refs: List[FileObjectRef],
        block_size: Optional[int] = None,
    ):
        address = remote_fileobj_refs[0].address
        client = await self._get_client(address)
        block_size = block_size or DEFAULT_TRANSFER_BLOCK_SIZE
        one_block_data = []
        current_file_size = 0
        for file_obj, remote_ref in zip(local_fileobjs, remote_fileobj_refs):
            while True:
                file_data = await file_obj.read(block_size)  # type: ignore
                if file_data:
                    one_block_data.append(
                        (remote_ref.address, remote_ref.uid, file_data)
                    )
                    current_file_size += len(file_data)
                    if current_file_size >= block_size:
                        message = self._gen_copy_to_fileobjs_message(one_block_data)
                        await self._call_with_client(client, message)
                        one_block_data.clear()
                        current_file_size = 0
                else:
                    break

        if current_file_size > 0:
            message = self._gen_copy_to_fileobjs_message(one_block_data)
            await self._call_with_client(client, message)
            one_block_data.clear()



# Copyright 2022-2025 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import os
import re
import shutil
import subprocess
import sys
import sysconfig
import tempfile
from importlib.metadata import distributions
from pathlib import Path
from typing import Optional

from packaging.requirements import Requirement
from packaging.version import Version

from .core import VirtualEnvManager
from .utils import is_vcs_url, run_subprocess_with_logger

UV_PATH = os.getenv("XOSCAR_UV_PATH")
SKIP_INSTALLED = bool(int(os.getenv("XOSCAR_VIRTUAL_ENV_SKIP_INSTALLED", "0")))
logger = logging.getLogger(__name__)


def _is_in_pyinstaller():
    return hasattr(sys, "_MEIPASS")


class UVVirtualEnvManager(VirtualEnvManager):
    def __init__(self, env_path: Path):
        super().__init__(env_path)
        self._install_process: Optional[subprocess.Popen] = None

    @classmethod
    def is_available(cls):
        if UV_PATH is not None:
            # user specified uv, just treat it as existed
            return True
        return shutil.which("uv") is not None

    @staticmethod
    def _get_uv_path() -> str:
        if (uv_path := UV_PATH) is None:
            try:
                from uv import find_uv_bin

                uv_path = find_uv_bin()
            except (ImportError, FileNotFoundError):
                logger.warning("Fail to find uv bin, use system one")
                uv_path = "uv"
        return uv_path

    def exists_env(self) -> bool:
        """Check if virtual environment already exists."""
        return self.env_path.exists() and (self.env_path / "pyvenv.cfg").exists()

    def create_env(
        self, python_path: Path | None = None, exists: str = "ignore"
    ) -> None:
        """
        Create virtual environment.

        Args:
            python_path: Path to Python interpreter to use
            exists: How to handle existing environment:
                - "ignore": Skip creation if environment already exists (default)
                - "error": Raise error if environment exists
                - "clear": Remove existing environment and create new one
        """
        if self.exists_env():
            if exists == "error":
                raise FileExistsError(
                    f"Virtual environment already exists at {self.env_path}"
                )
            elif exists == "ignore":
                logger.info(
                    f"Virtual environment already exists at {self.env_path}, skipping creation"
                )
                return
            elif exists == "clear":
                logger.info(f"Removing existing virtual environment at {self.env_path}")
                self.remove_env()
            else:
                raise ValueError(
                    f"Invalid exists option: {exists}. Must be one of: error, clear, ignore"
                )

        uv_path = self._get_uv_path()
        cmd = [uv_path, "venv", str(self.env_path), "--system-site-packages"]

        if python_path:
            cmd += ["--python", str(python_path)]
        elif _is_in_pyinstaller():
            # in pyinstaller, uv would find the system python
            # in this case we'd better specify the same python version
            python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
            cmd += ["--python", python_version]

        logger.info("Creating virtualenv via command: %s", cmd)
        subprocess.run(cmd, check=True)

    def _resolve_install_plan(
        self,
        specs: list[str],
        pinned: dict[str, str],
        index_url: str | None = None,
        extra_index_url: str | list[str] | None = None,
        index_strategy: str | None = None,
    ) -> list[str]:
        """
        Run uv --dry-run with pinned constraints and return
        a list like ['package==version', ...].
        """
        with tempfile.NamedTemporaryFile("w+", delete=True) as f:
            for name, ver in pinned.items():
                f.write(f"{name}=={ver}\n")
            f.flush()  # make sure content is on disk

            cmd = [
                self._get_uv_path(),
                "pip",
                "install",
                "-p",
                str(self.env_path),
                "--dry-run",
                "--constraint",
                f.name,
            ]

            # Add index URL parameters
            if index_url:
                cmd += ["-i", index_url]
            if extra_index_url:
                cmd += (
                    ["--extra-index-url", extra_index_url]
                    if isinstance(extra_index_url, str)
                    else [
                        opt for v in extra_index_url for opt in ("--extra-index-url", v)
                    ]
                )
            if index_strategy:
                cmd += ["--index-strategy", index_strategy]

            cmd.extend(specs)
            try:
                result = subprocess.run(cmd, check=True, text=True, capture_output=True)
            except subprocess.CalledProcessError as e:
                logger.error(e.stderr)
                raise e
        # the temp file is automatically deleted here
        deps = [
            f"{m.group(1)}=={m.group(2)}"
            for line in result.stderr.splitlines()
            if (m := re.match(r"^\+ (\S+)==(\S+)$", line.strip()))
        ]
        return deps

    @staticmethod
    def _split_specs(
        specs: list[str], installed: dict[str, str]
    ) -> tuple[list[str], list[str], dict[str, str]]:
        """
        Split the given requirement specs into:
        - keep： specs that need to be kept, e.g. git+github://xxx
        - to_resolve: specs that need to be passed to the resolver (unsatisfied ones)
        - pinned: already satisfied specs, used for constraint to lock their versions
        """
        keep: list[str] = []
        to_resolve: list[str] = []
        pinned: dict[str, str] = {}

        for spec_str in specs:
            # skip git+xxx
            if is_vcs_url(spec_str):
                keep.append(spec_str)
                continue

            req = Requirement(spec_str)
            name = req.name.lower()
            cur_ver = installed.get(name)

            if cur_ver is None:
                # Package not installed, needs resolution
                to_resolve.append(spec_str)
                continue

            if not req.specifier:
                # No version constraint, already satisfied
                pinned[name] = cur_ver
                continue

            try:
                if Version(cur_ver) in req.specifier:
                    # Version satisfies the specifier, pin it
                    pinned[name] = cur_ver
                else:
                    # Version does not satisfy, needs resolution
                    to_resolve.append(spec_str)
            except Exception:
                # Parsing error, be conservative and resolve it
                to_resolve.append(spec_str)

        return keep, to_resolve, pinned

    def _filter_packages_not_installed(
        self,
        packages: list[str],
        index_url: str | None = None,
        extra_index_url: str | list[str] | None = None,
        index_strategy: str | None = None,
    ) -> list[str]:
        """
        Filter out packages that are already installed with the same version.
        """

        # all the installed packages in system site packages
        installed = {
            dist.metadata["Name"].lower(): dist.version
            for dist in distributions()
            if dist.metadata and "Name" in dist.metadata
        }

        # exclude those packages that satisfied in system site packages
        keep, to_resolve, pinned = self._split_specs(packages, installed)
        if not keep and not to_resolve:
            logger.debug("All requirement specifiers satisfied by system packages.")
            return []

        if to_resolve:
            resolved = self._resolve_install_plan(
                to_resolve, pinned, index_url, extra_index_url, index_strategy
            )
            logger.debug(f"Resolved install list: {resolved}")
            if not keep and not resolved:
                # no packages to install
                return []
        else:
            resolved = []

        final = keep.copy()
        for item in resolved:
            name, version = item.split("==")
            key = name.lower()
            if key not in installed or installed[key] != version:
                final.append(item)
        logger.debug(f"Filtered install list: {final}")
        return final

    def install_packages(self, packages: list[str], **kwargs):
        """
        Install packages into the virtual environment using uv.

        Args:
            packages: List of package specifications
            **kwargs: Can include:
                - Pip configuration: index_url, extra_index_url, find_links, trusted_host,
                  no_build_isolation, log, skip_installed
                - Dynamic variables for #var# substitution: e.g., engine='vllm', mode='remote'
        """
        if not packages:
            return

        # Pop pip configuration parameters, remaining kwargs are for variable substitution
        log = kwargs.pop("log", False)
        skip_installed = kwargs.pop("skip_installed", SKIP_INSTALLED)
        index_url = kwargs.pop("index_url", None)
        extra_index_url = kwargs.pop("extra_index_url", None)
        index_strategy = kwargs.pop("index_strategy", None)

        # Process packages with variable substitution
        packages = self.process_packages(packages, **kwargs)
        if not packages:
            return

        uv_path = self._get_uv_path()

        if skip_installed:
            packages = self._filter_packages_not_installed(
                packages, index_url, extra_index_url, index_strategy
            )
            if not packages:
                logger.info("All required packages are already installed.")
                return

            cmd = [
                uv_path,
                "pip",
                "install",
                "-p",
                str(self.env_path),
                "--color=always",
                "--no-deps",
            ] + packages
        else:
            cmd = [
                uv_path,
                "pip",
                "install",
                "-p",
                str(self.env_path),
                "--color=always",
            ] + packages

        if index_url:
            cmd += ["-i", index_url]
        param_and_option = [
            (extra_index_url, "--extra-index-url"),
            ("find_links" in kwargs and kwargs["find_links"], "-f"),
            ("trusted_host" in kwargs and kwargs["trusted_host"], "--trusted-host"),
        ]
        for param, option in param_and_option:
            if param:
                val = (
                    param
                    if not isinstance(param, bool)
                    else kwargs.get(
                        {"-f": "find_links", "--trusted-host": "trusted_host"}[option]
                    )
                )
                if val:
                    cmd += (
                        [option, val]
                        if isinstance(val, str)
                        else [opt for v in val for opt in (option, v)]
                    )

        if index_strategy:
            cmd += ["--index-strategy", index_strategy]
        if kwargs.get("no_build_isolation", False):
            cmd += ["--no-build-isolation"]

        logger.info("Installing packages via command: %s", cmd)
        if not log:
            self._install_process = process = subprocess.Popen(cmd)
            returncode = process.wait()
        else:
            with run_subprocess_with_logger(cmd) as process:
                self._install_process = process
            returncode = process.returncode

        self._install_process = None
        if returncode != 0:
            raise subprocess.CalledProcessError(returncode, cmd)

    def cancel_install(self):
        if self._install_process and self._install_process.poll() is None:
            self._install_process.terminate()
            self._install_process.wait()

    def get_python_path(self) -> str | None:
        if self.env_path.exists():
            return str(self.env_path.joinpath("bin/python"))
        return None

    def get_lib_path(self) -> str:
        return sysconfig.get_path("purelib", vars={"base": str(self.env_path)})

    def remove_env(self):
        if self.env_path.exists():
            shutil.rmtree(self.env_path, ignore_errors=True)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import threading
from typing import Any, Optional, Type, Union

from .communication import Client, get_client_type

_CACHE_KEY_TYPE = Union[
    tuple[str, Any, Optional[Type[Client]]],
    tuple[str, Any, Optional[Type[Client]], Optional[tuple[str, ...]]],
]


class Router:
    """
    Router provides mapping from external address to internal address.
    """

    __slots__ = (
        "_curr_external_addresses",
        "_local_mapping",
        "_mapping",
        "_comm_config",
        "_proxy_config",
        "_cache_local",
    )

    _instance: "Router" | None = None

    @staticmethod
    def set_instance(router: Optional["Router"]):
        # Default router is set when an actor pool started
        Router._instance = router

    @staticmethod
    def get_instance() -> "Router" | None:
        return Router._instance

    @staticmethod
    def get_instance_or_empty() -> "Router":
        return Router._instance or Router(list(), None)

    def __init__(
        self,
        external_addresses: list[str],
        local_address: str | None,
        mapping: dict[str, str] | None = None,
        comm_config: dict | None = None,
        proxy_config: dict | None = None,
    ):
        self._curr_external_addresses = external_addresses
        self._local_mapping = dict()
        for addr in self._curr_external_addresses:
            self._local_mapping[addr] = local_address
        if mapping is None:
            mapping = dict()
        self._mapping = mapping
        self._comm_config = comm_config or dict()
        self._proxy_config = proxy_config or dict()
        self._cache_local = threading.local()

    @property
    def _cache(self) -> dict[_CACHE_KEY_TYPE, Client]:
        try:
            return self._cache_local.cache
        except AttributeError:
            cache = self._cache_local.cache = dict()
            return cache

    @property
    def _lock(self) -> asyncio.Lock:
        try:
            return self._cache_local.lock
        except AttributeError:
            lock = self._cache_local.lock = asyncio.Lock()
            return lock

    def set_mapping(self, mapping: dict[str, str]):
        self._mapping = mapping
        self._cache_local = threading.local()

    def add_router(self, router: "Router"):
        self._curr_external_addresses.extend(router._curr_external_addresses)
        self._local_mapping.update(router._local_mapping)
        self._mapping.update(router._mapping)
        self._comm_config.update(router._comm_config)
        self._proxy_config.update(router._proxy_config)
        self._cache_local = threading.local()

    def remove_router(self, router: "Router"):
        for external_address in router._curr_external_addresses:
            try:
                self._curr_external_addresses.remove(external_address)
            except ValueError:
                pass
        for addr in router._local_mapping:
            self._local_mapping.pop(addr, None)
        for addr in router._mapping:
            self._mapping.pop(addr, None)
        self._cache_local = threading.local()

    @property
    def external_address(self):
        if self._curr_external_addresses:
            return self._curr_external_addresses[0]

    def get_internal_address(self, external_address: str) -> str | None:
        try:
            # local address, use dummy address
            return self._local_mapping[external_address]
        except KeyError:
            # try to lookup inner address from address mapping
            return self._mapping.get(external_address)

    async def get_client(
        self,
        external_address: str,
        from_who: Any = None,
        cached: bool = True,
        proxy_addresses: list[str] | None = None,
        **kw,
    ) -> Client:
        async with self._lock:
            proxy_addrs: tuple[str, ...] | None = (
                tuple(proxy_addresses) if proxy_addresses else None
            )
            if (
                cached
                and (external_address, from_who, None, proxy_addrs) in self._cache
            ):
                cached_client = self._cache[
                    external_address, from_who, None, proxy_addrs
                ]
                if cached_client.closed:
                    # closed before, ignore it
                    del self._cache[external_address, from_who, None, proxy_addrs]
                else:
                    return cached_client

            address = self.get_internal_address(external_address)
            if address is None:
                # no inner address, just use external address
                address = external_address
                # check if proxy address exists
                proxy_address = proxy_addresses[-1] if proxy_addresses else None
                if proxy_address is None:
                    proxy_address = self.get_proxy(address)
                    if proxy_address and proxy_address != self.external_address:
                        address = proxy_address
                else:
                    if new_proxy_address := self.get_proxy(proxy_address):
                        address = new_proxy_address
                    else:
                        address = proxy_address

            client_type: Type[Client] = get_client_type(address)
            client = await self._create_client(client_type, address, **kw)
            if cached:
                self._cache[external_address, from_who, None, proxy_addrs] = client
            return client

    async def _create_client(
        self, client_type: Type[Client], address: str, **kw
    ) -> Client:
        config = client_type.parse_config(self._comm_config)
        if config:
            kw["config"] = config
        local_address = (
            self._curr_external_addresses[0] if self._curr_external_addresses else None
        )
        return await client_type.connect(address, local_address=local_address, **kw)

    def _get_client_type_to_addresses(
        self, external_address: str
    ) -> dict[Type[Client], str]:
        client_type_to_addresses = dict()
        client_type_to_addresses[get_client_type(external_address)] = external_address
        if external_address in self._curr_external_addresses:  # pragma: no cover
            # local address, use dummy address
            addr = self._local_mapping.get(external_address)
            client_type = get_client_type(addr)  # type: ignore
            client_type_to_addresses[client_type] = addr  # type: ignore
        if external_address in self._mapping:
            # try to lookup inner address from address mapping
            addr = self._mapping.get(external_address)
            client_type = get_client_type(addr)  # type: ignore
            client_type_to_addresses[client_type] = addr  # type: ignore
        return client_type_to_addresses

    def get_all_client_types(self, external_address: str) -> list[Type[Client]]:
        return list(self._get_client_type_to_addresses(external_address))

    async def get_client_via_type(
        self,
        external_address: str,
        client_type: Type[Client],
        from_who: Any = None,
        cached: bool = True,
        **kw,
    ) -> Client:
        async with self._lock:
            if cached and (external_address, from_who, client_type) in self._cache:
                cached_client = self._cache[external_address, from_who, client_type]
                if cached_client.closed:  # pragma: no cover
                    # closed before, ignore it
                    del self._cache[external_address, from_who, client_type]
                else:
                    return cached_client

            client_type_to_addresses = self._get_client_type_to_addresses(
                external_address
            )
            if client_type not in client_type_to_addresses:  # pragma: no cover
                raise ValueError(
                    f"Client type({client_type}) is not supported for {external_address}"
                )
            address = client_type_to_addresses[client_type]
            client = await self._create_client(client_type, address, **kw)
            if cached:
                self._cache[external_address, from_who, client_type] = client
            return client

    def get_proxy(self, from_addr: str) -> str | None:
        """
        Get proxy address that sent to.

        Some patterns can be supported:

        1. Direct address mapping, e.g. mapping 127.0.0.1:12345 -> 127.0.0.1:12346
           The message will be sent to 127.0.0.1:12346 as forward one.
        2. Host match, e.g. mapping 127.0.0.1 -> 127.0.0.1:12346
           All the messages that match the host, e.g. 127.0.0.1:12345 and 127.0.0.1:12347
           will be sent to 127.0.0.1:12346 as forward one.
        3. Wildcard, e.g. mapping * -> 127.0.0.1:12346
           All the messages will be sent to 127.0.0.1:12346 as forward one.
        """

        host = from_addr.split(":", 1)[0]

        proxy_map = self._proxy_config
        addr = proxy_map.get(from_addr)
        if addr and addr != from_addr:
            return addr
        addr = proxy_map.get(host)
        if addr and addr != from_addr:
            return addr
        addr = proxy_map.get("*")
        if addr and addr != from_addr:
            return addr
        return None

    def get_proxies(self, from_addr: str) -> list[str] | None:
        """
        Get all proxies

        e.g. Proxy mapping {'a': 'b', 'b': 'c'}
        get_proxies('a') will return ['b', 'c']
        """

        proxies: list[str] = []
        while True:
            proxy = self.get_proxy(from_addr)
            if not proxies and not proxy:
                return None
            elif not proxy:
                return proxies
            else:
                proxies.append(proxy)
                from_addr = proxy



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, List, Optional, Union
from urllib.parse import urlparse

from ._utils cimport new_actor_id, new_random_id
from .core cimport ActorRef, BufferRef, FileObjectRef


cdef dict _backend_context_cls = dict()

cdef object _context = None


cdef class BaseActorContext:
    # allocate strategy is for Indigen backend only
    support_allocate_strategy = False

    """
    Base class for actor context. Every backend need to implement
    actor context for their own.
    """

    def __init__(self, address: str = None):
        self._address = address

    async def create_actor(
        self,
        object actor_cls,
        *args,
        object uid=None,
        object address=None,
        **kwargs,
    ):
        """
        Stub method for creating an actor in current context.

        Parameters
        ----------
        actor_cls : Actor
            Actor class
        args : tuple
            args to be passed into actor_cls.__init__
        uid : identifier
            Actor identifier
        address : str
            Address to locate the actor
        kwargs : dict
            kwargs to be passed into actor_cls.__init__

        Returns
        -------
        ActorRef

        """
        raise NotImplementedError

    async def has_actor(self, ActorRef actor_ref):
        """
        Check if actor exists in current context

        Parameters
        ----------
        actor_ref : ActorRef
            Reference to an actor

        Returns
        -------
        bool
        """
        raise NotImplementedError

    async def destroy_actor(self, ActorRef actor_ref):
        """
        Destroy an actor by its reference

        Parameters
        ----------
        actor_ref : ActorRef
            Reference to an actor

        Returns
        -------
        bool
        """
        raise NotImplementedError

    async def kill_actor(self, ActorRef actor_ref):
        """
        Force to kill an actor, take care this is a dangerous operation,
        it may lead to the result that other actors are killed as well.
        Hence, unless you are knowing what you are doing and know how
        to recover possible effected actors, DO NOT USE this method!

        Parameters
        ----------
        actor_ref : ActorRef
            Reference to an actor

        Returns
        -------
        bool
        """

    async def send(
        self,
        ActorRef actor_ref,
        object message,
        bint wait_response=True,
        object profiling_context=None,
    ):
        """
        Send a message to given actor by its reference

        Parameters
        ----------
        actor_ref : ActorRef
            Reference to an actor
        message : object
            Message to send to an actor, need to comply to Actor.__on_receive__
        wait_response : bool
            Whether to wait for responses from the actor.
        profiling_context: ProfilingContext
            The profiling context.

        Returns
        -------
        object
        """
        raise NotImplementedError

    async def actor_ref(self, *args, **kwargs):
        """
        Create a reference to an actor

        Returns
        -------
        ActorRef
        """
        raise NotImplementedError

    async def wait_actor_pool_recovered(self, str address, str main_address = None):
        """
        Wait until an actor pool is recovered

        Parameters
        ----------
        address
            address of the actor pool
        main_address
            address of the main pool
        """
        raise NotImplementedError

    async def get_pool_config(self, str address):
        """
        Get config of actor pool with given address

        Parameters
        ----------
        address
            address of the actor pool

        Returns
        -------

        """
        raise NotImplementedError

    def buffer_ref(self, str address, object buf) -> BufferRef:
        """
        Create a reference to a buffer

        Parameters
        ----------
        address
            address of the actor pool
        buf
            buffer object

        Returns
        -------
        BufferRef
        """
        return BufferRef.create(buf, address, new_random_id(32))

    def file_object_ref(self, str address, object file_object) -> FileObjectRef:
        """
        Create a reference to an aio file object

        Parameters
        ----------
        address
            address of the actor pool
        file_object
            aio file object

        Returns
        -------
        FileObjectRef
        """
        return FileObjectRef.create(file_object, address, new_random_id(32))

    async def copy_to_buffers(self, local_buffers: List, remote_buffer_refs: List[BufferRef], block_size: Optional[int] = None):
        """
        Copy local buffers to remote buffers.
        Parameters
        ----------
        local_buffers
            Local buffers.
        remote_buffer_refs
            Remote buffer refs
        block_size
            Transfer block size when non-ucx
        """
        raise NotImplementedError

    async def copy_to_fileobjs(self, local_fileobjs: list, remote_fileobj_refs: List[FileObjectRef], block_size: Optional[int] = None):
        """
        Copy local file objs to remote file objs.
        Parameters
        ----------
        local_fileobjs
            Local file objs.
        remote_fileobj_refs
            Remote file object refs
        block_size
            Transfer block size when non-ucx
        """
        raise NotImplementedError


cdef class ClientActorContext(BaseActorContext):
    """
    Default actor context. This context will keep references to other contexts
    given their protocol scheme (i.e., `ray://xxx`).
    """
    cdef dict _backend_contexts

    def __init__(self, address: str = None):
        BaseActorContext.__init__(self, address)
        self._backend_contexts = dict()

    cdef inline object _get_backend_context(self, object address):
        if address is None:
            raise ValueError('address has to be provided')
        if '://' not in address:
            scheme = None
        else:
            scheme = urlparse(address).scheme or None
        try:
            return self._backend_contexts[scheme]
        except KeyError:
            context = self._backend_contexts[scheme] = \
                _backend_context_cls[scheme](address)
            return context

    def create_actor(
        self,
        object actor_cls,
        *args,
        object uid=None,
        object address=None,
        **kwargs,
    ):
        context = self._get_backend_context(address)
        uid = uid or new_actor_id()
        return context.create_actor(actor_cls, *args, uid=uid, address=address, **kwargs)

    def has_actor(self, ActorRef actor_ref):
        context = self._get_backend_context(actor_ref.address)
        return context.has_actor(actor_ref)

    def destroy_actor(self, ActorRef actor_ref):
        context = self._get_backend_context(actor_ref.address)
        return context.destroy_actor(actor_ref)

    def kill_actor(self, ActorRef actor_ref):
        context = self._get_backend_context(actor_ref.address)
        return context.kill_actor(actor_ref)

    def actor_ref(self, *args, **kwargs):
        from ._utils import create_actor_ref

        actor_ref = create_actor_ref(*args, **kwargs)
        context = self._get_backend_context(actor_ref.address)
        return context.actor_ref(actor_ref)

    def send(
        self,
        ActorRef actor_ref,
        object message,
        bint wait_response=True,
        object profiling_context=None
    ):
        context = self._get_backend_context(actor_ref.address)
        return context.send(
            actor_ref,
            message,
            wait_response=wait_response,
            profiling_context=profiling_context,
        )

    def wait_actor_pool_recovered(self, str address, str main_address = None):
        context = self._get_backend_context(address)
        return context.wait_actor_pool_recovered(address, main_address)

    def get_pool_config(self, str address):
        context = self._get_backend_context(address)
        return context.get_pool_config(address)

    def buffer_ref(self, str address, buf: Any) -> BufferRef:
        context = self._get_backend_context(address)
        return context.buffer_ref(address, buf)

    def file_object_ref(self, str address, object file_object) -> FileObjectRef:
        context = self._get_backend_context(address)
        return context.file_object_ref(address, file_object)

    def copy_to(self, local_buffers_or_fileobjs: list, remote_refs: List[Union[BufferRef, FileObjectRef]], block_size: Optional[int] = None):
        if len(local_buffers_or_fileobjs) == 0 or len(remote_refs) == 0:
            raise ValueError("Nothing to transfer since the length of `local_buffers_or_fileobjs` or `remote_refs` is 0.")
        assert (
            len({ref.address for ref in remote_refs}) == 1
        ), "remote_refs for `copy_to` can support only 1 destination"
        assert len(local_buffers_or_fileobjs) == len(remote_refs), (
            f"Buffers or fileobjs from local and remote must have same size, "
            f"local: {len(local_buffers_or_fileobjs)}, remote: {len(remote_refs)}"
        )
        if block_size is not None:
            assert (
                block_size > 0
            ), f"`block_size` option must be greater than 0, current value: {block_size}."
        remote_ref = remote_refs[0]
        address = remote_ref.address
        context = self._get_backend_context(address)
        if isinstance(remote_ref, BufferRef):
            return context.copy_to_buffers(local_buffers_or_fileobjs, remote_refs, block_size)
        else:
            return context.copy_to_fileobjs(local_buffers_or_fileobjs, remote_refs, block_size)


def register_backend_context(scheme, cls):
    assert issubclass(cls, BaseActorContext)
    _backend_context_cls[scheme] = cls


cpdef get_context():
    """
    Get an actor context. If not in an actor environment,
    ClientActorContext will be used
    """
    global _context
    if _context is None:
        _context = ClientActorContext()
    return _context



# Copyright 2024 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Tuple

import numpy as np

from ..utils import lazy_import
from .core import Serializer, buffered

# lazy import PyTorch to avoid enforced dependency
torch = lazy_import("torch")
cupy = lazy_import("cupy")


def rmm_to_torch(buf):
    cupy_arr = cupy.asarray(buf)  # zero-copy
    torch_tensor = torch.utils.dlpack.from_dlpack(cupy_arr.toDlpack())  # zero-copy
    return torch_tensor


class TorchTensorSerializer(Serializer):
    @buffered
    def serial(self, obj: "torch.Tensor", context: Dict):  # type: ignore
        # for cpu tensor, use memory viewpoint
        if obj.device.type == "cpu":
            # make sure tensor is contiguous
            if not obj.is_contiguous():
                obj = obj.contiguous()
            # get memory viewpoint and collect header information
            header = {
                "shape": tuple(obj.shape),
                "dtype": str(obj.dtype),
                "device": obj.device.type,
                "requires_grad": obj.requires_grad,
                "strides": tuple(obj.stride()),
            }
            # Try zero-copy path: torch -> numpy
            try:
                np_array = obj.numpy()  # zero-copy if supported
                buffer = memoryview(np_array)
                header["format"] = "numpy"
                return (header,), [buffer], True

            except Exception:
                # Fallback: copy to uint8 bytes
                # This works for ANY dtype
                byte_tensor = obj.view(torch.uint8).clone()  # explicit copy
                np_array = byte_tensor.numpy()
                buffer = memoryview(np_array)
                header["format"] = "bytes"
                return (header,), [buffer], True
        elif obj.device.type == "cuda":
            # for CUDA， use __cuda_array_interface__
            if not (
                obj.is_contiguous()
                or obj.is_contiguous(memory_format=torch.channels_last)
            ):
                obj = obj.contiguous()

            # get cuda array interface information
            header = {
                "shape": tuple(obj.shape),
                "dtype": str(obj.dtype),
                "device": obj.device.type,
                "device_index": obj.device.index,
                "requires_grad": obj.requires_grad,
                "strides": tuple(obj.stride()),
            }

            # ---- Core idea: expose raw CUDA memory as a uint8 buffer (zero-copy) ----
            # Get the underlying untyped storage that actually owns the CUDA memory
            storage = obj.untyped_storage()

            # Create a uint8 CUDA tensor that will act as a byte-view of the same memory
            # This does NOT allocate new GPU memory; it only creates a new Tensor wrapper
            buffer = torch.empty(
                (storage.nbytes(),),
                dtype=torch.uint8,
                device=obj.device,
            )

            # Make the buffer tensor share the same storage (zero-copy)
            buffer.set_(storage)  # type: ignore[attr-defined]

            # Return: (metadata,), [raw CUDA buffer], mark as buffered
            return (header,), [buffer], True
        else:
            # for unsupported device
            raise NotImplementedError(f"Unsupported device type: {obj.device.type}")

    def deserial(self, serialized: Tuple, context: Dict, subs: List[Any]):
        header = serialized[0]
        device = header["device"]
        data_buffer = subs[0]

        if device == "cpu":
            fmt = header.get("format", "numpy")
            dtype_name = header["dtype"].split(".")[-1]
            dtype = getattr(torch, dtype_name)

            if fmt == "numpy":
                # zero-copy path
                np_array = np.frombuffer(data_buffer, dtype=np.dtype(dtype_name))
                tensor = torch.from_numpy(np_array).view(header["shape"])

            else:
                # bytes path (copy unavoidable)
                byte_np = np.frombuffer(data_buffer, dtype=np.uint8)
                byte_tensor = torch.from_numpy(byte_np)

                tensor = torch.empty(0, dtype=dtype)
                tensor = tensor.set_(
                    byte_tensor.untyped_storage(),
                    storage_offset=0,
                    size=tuple(header["shape"]),
                    stride=tuple(header["strides"]),
                )

            if header.get("requires_grad"):
                tensor.requires_grad_(True)

        elif device == "cuda":
            # Unpack metadata
            (header,) = serialized

            # Raw CUDA buffer (uint8 tensor)
            buffer = subs[0]
            if not isinstance(buffer, torch.Tensor):
                buffer = rmm_to_torch(buffer)
            assert buffer.is_cuda, "buffer must be a CUDA tensor"
            assert buffer.dtype == torch.uint8, "buffer must be uint8"

            # Get the shared CUDA storage
            storage = buffer.untyped_storage()

            # Resolve original dtype
            dtype_name = header["dtype"].split(".")[-1]
            dtype = getattr(torch, dtype_name)

            # Create an empty tensor wrapper with the correct dtype
            # This does NOT allocate new GPU memory for data
            tensor = torch.empty(0, device=buffer.device, dtype=dtype)

            # Bind the tensor to the same storage with original shape/stride
            # Pure zero-copy: only a new Tensor view, no data movement
            tensor.set_(
                storage,
                storage_offset=0,
                size=tuple(header["shape"]),
                stride=tuple(header["strides"]),
            )

            # Restore requires_grad if needed
            if header.get("requires_grad"):
                tensor.requires_grad_(True)

            return tensor
        else:
            raise NotImplementedError(f"Unsupported device type: {device}")

        # recover requires_grad attributes
        tensor.requires_grad = header["requires_grad"]
        return tensor


# only when torch is available, we register module
if torch is not None:
    TorchTensorSerializer.register("torch.Tensor")



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import IntEnum
from typing import Dict, Type

import numpy as np

from ..utils import lazy_import
from . import xoscar_pygloo as xp

ReduceOpMappingGloo: Dict["CollectiveReduceOp", "xp.ReduceOp"] = {}
AllReduceAlgorithmMappingGloo: Dict["AllReduceAlgorithm", "xp.AllreduceAlgorithm"] = {}


def _register_reduce_op(reduce_op):
    for op_type in reduce_op:
        ReduceOpMappingGloo[op_type] = xp.ReduceOp(op_type)
    return reduce_op


def _register_allreduce_algo(algorithms):
    for algo in algorithms:
        AllReduceAlgorithmMappingGloo[algo] = xp.AllreduceAlgorithm(algo)
    return algorithms


@_register_reduce_op
class CollectiveReduceOp(IntEnum):
    SUM = 0
    PRODUCT = 1
    MIN = 2
    MAX = 3
    BAND = 4
    BOR = 5
    BXOR = 6
    UNUSED = 7


@_register_allreduce_algo
class AllReduceAlgorithm(IntEnum):
    UNSPECIFIED = 0
    RING = 1
    BCUBE = 2


TypeMappingGloo: Dict[Type[np.dtype], "xp.GlooDataType_t"] = {
    np.int8: xp.GlooDataType_t.glooInt8,  # type: ignore
    np.uint8: xp.GlooDataType_t.glooUint8,  # type: ignore
    np.int32: xp.GlooDataType_t.glooInt32,  # type: ignore
    np.uint32: xp.GlooDataType_t.glooUint32,  # type: ignore
    np.int64: xp.GlooDataType_t.glooInt64,  # type: ignore
    np.uint64: xp.GlooDataType_t.glooUint64,  # type: ignore
    np.float16: xp.GlooDataType_t.glooFloat16,  # type: ignore
    np.float32: xp.GlooDataType_t.glooFloat32,  # type: ignore
    np.float64: xp.GlooDataType_t.glooFloat64,  # type: ignore
}
cupy = lazy_import("cupy")
if cupy is not None:
    from cupy.cuda import nccl

    TypeMappingNCCL: Dict[Type[np.dtype], int] = {
        np.int8: nccl.NCCL_INT8,  # type: ignore
        np.uint8: nccl.NCCL_UINT8,  # type: ignore
        np.int32: nccl.NCCL_INT32,  # type: ignore
        np.uint32: nccl.NCCL_UINT32,  # type: ignore
        np.int64: nccl.NCCL_INT64,  # type: ignore
        np.uint64: nccl.NCCL_UINT64,  # type: ignore
        np.float16: nccl.NCCL_FLOAT16,  # type: ignore
        np.float32: nccl.NCCL_FLOAT32,  # type: ignore
        np.float64: nccl.NCCL_FLOAT64,  # type: ignore
    }

    ReduceOpMappingNCCL: Dict[CollectiveReduceOp, int] = {
        CollectiveReduceOp.SUM: nccl.NCCL_SUM,
        CollectiveReduceOp.PRODUCT: nccl.NCCL_PROD,
        CollectiveReduceOp.MAX: nccl.NCCL_MAX,
        CollectiveReduceOp.MIN: nccl.NCCL_MIN,
    }

    ReduceOpMappingNCCLStr: Dict[CollectiveReduceOp, str] = {
        CollectiveReduceOp.SUM: "sum",
        CollectiveReduceOp.PRODUCT: "prod",
        CollectiveReduceOp.MAX: "max",
        CollectiveReduceOp.MIN: "min",
    }
# Some static variables
INVOKE_ERROR_MESSAGE = "Collective-related functions must be called in a process that is involved in collection communication."
RANK_ADDRESS_ENV_KEY = "COLLECTIVE_RANK_ADDRESS"
RENDEZVOUS_MASTER_IP_ENV_KEY = "COLLECTIVE_MASTER_IP"
RENDEZVOUS_MASTER_PORT_ENV_KEY = "COLLECTIVE_MASTER_PORT"
COLLECTIVE_DEVICE_ID_ENV_KEY = "COLLECTIVE_DEVICE_ID_FOR_AN_ACTOR"



# Copyright 2017 The Ray Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import subprocess
import sys

import psutil

# Linux can bind child processes' lifetimes to that of their parents via prctl.
# prctl support is detected dynamically once, and assumed thereafter.
linux_prctl = None

# Windows can bind processes' lifetimes to that of kernel-level "job objects".
# We keep a global job object to tie its lifetime to that of our own process.
win32_job = None
win32_AssignProcessToJobObject = None

logger = logging.getLogger(__name__)


def detect_fate_sharing_support_win32():
    global win32_job, win32_AssignProcessToJobObject
    if win32_job is None and sys.platform == "win32":
        import ctypes

        try:
            from ctypes.wintypes import BOOL, DWORD, HANDLE, LPCWSTR, LPVOID

            kernel32 = ctypes.WinDLL("kernel32")
            kernel32.CreateJobObjectW.argtypes = (LPVOID, LPCWSTR)
            kernel32.CreateJobObjectW.restype = HANDLE
            sijo_argtypes = (HANDLE, ctypes.c_int, LPVOID, DWORD)
            kernel32.SetInformationJobObject.argtypes = sijo_argtypes
            kernel32.SetInformationJobObject.restype = BOOL
            kernel32.AssignProcessToJobObject.argtypes = (HANDLE, HANDLE)
            kernel32.AssignProcessToJobObject.restype = BOOL
            kernel32.IsDebuggerPresent.argtypes = ()
            kernel32.IsDebuggerPresent.restype = BOOL
        except (AttributeError, TypeError, ImportError):
            kernel32 = None
        job = kernel32.CreateJobObjectW(None, None) if kernel32 else None
        job = subprocess.Handle(job) if job else job
        if job:
            from ctypes.wintypes import DWORD, LARGE_INTEGER, ULARGE_INTEGER

            class JOBOBJECT_BASIC_LIMIT_INFORMATION(ctypes.Structure):
                _fields_ = [
                    ("PerProcessUserTimeLimit", LARGE_INTEGER),
                    ("PerJobUserTimeLimit", LARGE_INTEGER),
                    ("LimitFlags", DWORD),
                    ("MinimumWorkingSetSize", ctypes.c_size_t),
                    ("MaximumWorkingSetSize", ctypes.c_size_t),
                    ("ActiveProcessLimit", DWORD),
                    ("Affinity", ctypes.c_size_t),
                    ("PriorityClass", DWORD),
                    ("SchedulingClass", DWORD),
                ]

            class IO_COUNTERS(ctypes.Structure):
                _fields_ = [
                    ("ReadOperationCount", ULARGE_INTEGER),
                    ("WriteOperationCount", ULARGE_INTEGER),
                    ("OtherOperationCount", ULARGE_INTEGER),
                    ("ReadTransferCount", ULARGE_INTEGER),
                    ("WriteTransferCount", ULARGE_INTEGER),
                    ("OtherTransferCount", ULARGE_INTEGER),
                ]

            class JOBOBJECT_EXTENDED_LIMIT_INFORMATION(ctypes.Structure):
                _fields_ = [
                    ("BasicLimitInformation", JOBOBJECT_BASIC_LIMIT_INFORMATION),
                    ("IoInfo", IO_COUNTERS),
                    ("ProcessMemoryLimit", ctypes.c_size_t),
                    ("JobMemoryLimit", ctypes.c_size_t),
                    ("PeakProcessMemoryUsed", ctypes.c_size_t),
                    ("PeakJobMemoryUsed", ctypes.c_size_t),
                ]

            debug = kernel32.IsDebuggerPresent()

            # Defined in <WinNT.h>; also available here:
            # https://docs.microsoft.com/en-us/windows/win32/api/jobapi2/nf-jobapi2-setinformationjobobject
            JobObjectExtendedLimitInformation = 9
            JOB_OBJECT_LIMIT_BREAKAWAY_OK = 0x00000800
            JOB_OBJECT_LIMIT_DIE_ON_UNHANDLED_EXCEPTION = 0x00000400
            JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE = 0x00002000
            buf = JOBOBJECT_EXTENDED_LIMIT_INFORMATION()
            buf.BasicLimitInformation.LimitFlags = (
                (0 if debug else JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE)
                | JOB_OBJECT_LIMIT_DIE_ON_UNHANDLED_EXCEPTION
                | JOB_OBJECT_LIMIT_BREAKAWAY_OK
            )
            infoclass = JobObjectExtendedLimitInformation
            if not kernel32.SetInformationJobObject(
                job, infoclass, ctypes.byref(buf), ctypes.sizeof(buf)
            ):
                job = None
        win32_AssignProcessToJobObject = (
            kernel32.AssignProcessToJobObject if kernel32 is not None else False
        )
        win32_job = job if job else False
    return bool(win32_job)


def detect_fate_sharing_support_linux():
    global linux_prctl
    if linux_prctl is None and sys.platform.startswith("linux"):
        try:
            from ctypes import CDLL, c_int, c_ulong

            prctl = CDLL(None).prctl
            prctl.restype = c_int
            prctl.argtypes = [c_int, c_ulong, c_ulong, c_ulong, c_ulong]
        except (AttributeError, TypeError):
            prctl = None
        linux_prctl = prctl if prctl else False
    return bool(linux_prctl)


def detect_fate_sharing_support():
    result = None
    if sys.platform == "win32":
        result = detect_fate_sharing_support_win32()
    elif sys.platform.startswith("linux"):
        result = detect_fate_sharing_support_linux()
    return result


if detect_fate_sharing_support():
    logger.info("Using kernel-level fate-sharing.")


def set_kill_on_parent_death_linux():
    """Ensures this process dies if its parent dies (fate-sharing).

    Linux-only. Must be called in preexec_fn (i.e. by the child).
    """
    if detect_fate_sharing_support_linux():
        import signal

        PR_SET_PDEATHSIG = 1
        if linux_prctl(PR_SET_PDEATHSIG, signal.SIGKILL, 0, 0, 0) != 0:
            import ctypes

            raise OSError(ctypes.get_errno(), "prctl(PR_SET_PDEATHSIG) failed")
    else:
        assert False, "PR_SET_PDEATHSIG used despite being unavailable"


def set_kill_child_on_death_win32(child_proc):
    """Ensures the child process dies if this process dies (fate-sharing).

    Windows-only. Must be called by the parent, after spawning the child.

    Args:
        child_proc: The subprocess.Popen or subprocess.Handle object.
    """

    if isinstance(child_proc, subprocess.Popen):
        child_proc = child_proc._handle
    assert isinstance(child_proc, subprocess.Handle)

    if detect_fate_sharing_support_win32():
        if not win32_AssignProcessToJobObject(win32_job, int(child_proc)):
            import ctypes

            raise OSError(ctypes.get_last_error(), "AssignProcessToJobObject() failed")
    else:
        assert False, "AssignProcessToJobObject used despite being unavailable"


def preexec_fn():
    import signal

    signal.pthread_sigmask(signal.SIG_BLOCK, {signal.SIGINT})
    if sys.platform.startswith("linux"):
        set_kill_on_parent_death_linux()


async def create_subprocess_exec(*args, **kwargs):
    win32 = sys.platform == "win32"
    # With Windows fate-sharing, we need special care:
    # The process must be added to the job before it is allowed to execute.
    # Otherwise, there's a race condition: the process might spawn children
    # before the process itself is assigned to the job.
    # After that point, its children will not be added to the job anymore.
    CREATE_SUSPENDED = 0x00000004  # from Windows headers
    creationflags = CREATE_SUSPENDED if win32 else 0
    if win32 and detect_fate_sharing_support_win32():
        creationflags |= subprocess.CREATE_NEW_PROCESS_GROUP
    # CREATE_NEW_PROCESS_GROUP is used to send Ctrl+C on Windows:
    # https://docs.python.org/3/library/subprocess.html#subprocess.Popen.send_signal
    process: asyncio.subprocess.Process = await asyncio.create_subprocess_exec(
        *args,
        **kwargs,
        preexec_fn=preexec_fn if not win32 else None,
        creationflags=creationflags,
    )
    if win32:
        proc = process._transport._proc
        try:
            set_kill_child_on_death_win32(proc)
            psutil.Process(process.pid).resume()
        except (psutil.Error, OSError):
            logger.exception("Resume process failed, kill %s.", process.pid)
            process.kill()
            raise
    return process



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asyncio import StreamReader, StreamWriter
from typing import Dict, List, Union

import numpy as np

from ...serialization.aio import BUFFER_SIZES_NAME
from ...utils import lazy_import

cupy = lazy_import("cupy")
cudf = lazy_import("cudf")
rmm = lazy_import("rmm")

CUDA_CHUNK_SIZE = 16 * 1024**2


def _convert_to_cupy_ndarray(
    cuda_buffer: Union["cupy.ndarray", "rmm.DeviceBuffer"]  # type: ignore
) -> "cupy.ndarray":  # type: ignore
    if isinstance(cuda_buffer, cupy.ndarray):
        return cuda_buffer

    size = cuda_buffer.nbytes
    data = cuda_buffer.__cuda_array_interface__["data"][0]
    memory = cupy.cuda.UnownedMemory(data, size, cuda_buffer)
    ptr = cupy.cuda.MemoryPointer(memory, 0)
    return cupy.ndarray(shape=size, dtype="u1", memptr=ptr)


def write_buffers(writer: StreamWriter, buffers: List):
    def _write_cuda_buffer(cuda_buffer: Union["cupy.ndarray", "rmm.DeviceBuffer"]):  # type: ignore
        # convert cuda buffer to cupy ndarray
        cuda_buffer = _convert_to_cupy_ndarray(cuda_buffer)

        chunk_size = CUDA_CHUNK_SIZE
        offset = 0
        nbytes = buffer.nbytes
        while offset < nbytes:
            size = chunk_size if (offset + chunk_size) < nbytes else nbytes - offset
            # slice on cupy ndarray
            chunk_buffer = cuda_buffer[offset : offset + size]
            # `get` will return numpy ndarray,
            # write its data which is a memoryview into writer
            writer.write(chunk_buffer.get().data)
            offset += size

    for buffer in buffers:
        if hasattr(buffer, "__cuda_array_interface__"):
            # GPU buffer
            _write_cuda_buffer(buffer)
        else:
            writer.write(buffer)


async def read_buffers(header: Dict, reader: StreamReader):
    is_cuda_buffers = header[0].get("is_cuda_buffers")
    buffer_sizes = header[0].pop(BUFFER_SIZES_NAME)

    buffers = []
    for is_cuda_buffer, buf_size in zip(is_cuda_buffers, buffer_sizes):
        if is_cuda_buffer:  # pragma: no cover
            if buf_size == 0:
                # uniformly use rmm.DeviceBuffer for cuda's deserialization
                buffers.append(rmm.DeviceBuffer(size=buf_size))
            else:
                buffer = rmm.DeviceBuffer(size=buf_size)
                arr = _convert_to_cupy_ndarray(buffer)
                offset = 0
                chunk_size = CUDA_CHUNK_SIZE
                while offset < buf_size:
                    read_size = (
                        chunk_size
                        if (offset + chunk_size) < buf_size
                        else buf_size - offset
                    )
                    content = await reader.readexactly(read_size)
                    chunk_arr = np.frombuffer(content, dtype="u1")
                    arr[offset : offset + len(content)].set(chunk_arr)
                    offset += read_size
                buffers.append(buffer)
        else:
            buffers.append(await reader.readexactly(buf_size))
    return buffers



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from enum import Enum
from types import TracebackType
from typing import Any, Type

from tblib import pickling_support

from ..core cimport ActorRef, BufferRef
from ..serialization.core cimport Serializer

from ..utils import wrap_exception

from .._utils cimport new_random_id

# make sure traceback can be pickled
pickling_support.install()

cdef int _DEFAULT_PROTOCOL = 0
DEFAULT_PROTOCOL = _DEFAULT_PROTOCOL


class MessageType(Enum):
    control = 0
    result = 1
    error = 2
    create_actor = 3
    destroy_actor = 4
    has_actor = 5
    actor_ref = 6
    send = 7
    tell = 8
    cancel = 9
    copy_to_buffers = 10
    copy_to_fileobjs = 11
    forward = 12


class ControlMessageType(Enum):
    stop = 0
    restart = 1
    sync_config = 2
    get_config = 3
    wait_pool_recovered = 4
    add_sub_pool_actor = 5
    # the new channel created is for data transfer only
    switch_to_copy_to = 6


cdef class _MessageSerialItem:
    cdef:
        tuple serialized
        list subs

    def __cinit__(self, tuple serialized, list subs):
        self.serialized = serialized
        self.subs = subs


cdef class _MessageBase:
    message_type: MessageType = None

    cdef:
        public int protocol
        public bytes message_id
        public list message_trace
        public object profiling_context

    def __init__(
        self,
        bytes message_id = None,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
        object profiling_context = None,
    ):
        self.message_id = message_id
        self.protocol = protocol
        # A message can be in the scope of other messages,
        # this is mainly used for detecting deadlocks,
        # e.g. Actor `A` sent a message(id: 1) to actor `B`,
        # in the processing of `B`, it sent back a message(id: 2) to `A`,
        # deadlock happens, because `A` is still waiting for reply from `B`.
        # In this case, the `scoped_message_ids` will be [1, 2],
        # `A` will find that id:1 already exists in inbox,
        # thus deadlock detected.
        self.message_trace = message_trace
        self.profiling_context = profiling_context

    cdef _MessageSerialItem serial(self):
        return _MessageSerialItem(
            (
                self.message_type.value,
                self.message_id,
                self.protocol,
                self.message_trace,
                self.profiling_context,
            ),
            [],
        )

    cdef deserial_members(self, tuple serialized, list subs):
        self.message_id = serialized[1]
        self.protocol = serialized[2]
        self.message_trace = serialized[3]
        self.profiling_context = serialized[4]

    def __repr__(self):
        cdef list attr_reprs = []
        for attr in dir(self):
            if attr.startswith("_") or attr == "message_type":
                continue
            val = getattr(self, attr)
            if callable(val):
                continue
            attr_reprs.append(f"{attr}={val!r}")
        values = ", ".join(attr_reprs)
        return f"{type(self).__name__}({values})"


cdef class ControlMessage(_MessageBase):
    message_type = MessageType.control

    cdef:
        public str address
        public object control_message_type
        public object content

    def __init__(
        self,
        bytes message_id = None,
        str address = None,
        object control_message_type: ControlMessageType = None,
        object content: Any = None,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
    ):
        _MessageBase.__init__(
            self, message_id, protocol=protocol, message_trace=message_trace
        )
        self.address = address
        self.control_message_type = control_message_type
        self.content = content

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        item.serialized += (
            self.address,
            self.control_message_type,
        )
        item.subs = [self.content]
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        _MessageBase.deserial_members(self, serialized, subs)
        self.address = serialized[-2]
        self.control_message_type = serialized[-1]
        self.content = subs[0]


cdef class ResultMessage(_MessageBase):
    message_type = MessageType.result

    cdef:
        public object result

    def __init__(
        self,
        bytes message_id = None,
        object result: Any = None,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
        object profiling_context = None,
    ):
        _MessageBase.__init__(
            self,
            message_id,
            protocol=protocol,
            message_trace=message_trace,
            profiling_context=profiling_context,
        )
        self.result = result

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        item.subs = [self.result]
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        _MessageBase.deserial_members(self, serialized, subs)
        self.result = subs[0]


class _AsCauseBase:
    def __str__(self):
        return f"[address={self.address}, pid={self.pid}] {str(self.__wrapped__)}"


cdef class ErrorMessage(_MessageBase):
    message_type = MessageType.error

    cdef:
        public str address
        public long pid
        public type error_type
        public object error
        public object traceback

    def __init__(
        self,
        bytes message_id = None,
        str address: str = None,
        long pid = -1,
        type error_type: Type[BaseException] = None,
        object error: BaseException = None,
        object traceback: TracebackType = None,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
    ):
        _MessageBase.__init__(
            self, message_id, protocol=protocol, message_trace=message_trace
        )
        self.address = address
        self.pid = pid
        self.error_type = error_type
        self.error = error
        self.traceback = traceback

    def as_instanceof_cause(self):
        # Check the as_instanceof_cause is not recursive.
        #
        # e.g. actor_a.method1 will reraise the exception raised
        # from actor_b.method2. But these two actors are in the same
        # process, so we don't want to append duplicated address and pid in the
        # error message.
        if issubclass(self.error_type, _AsCauseBase):
            return self.error.with_traceback(self.traceback)

        # for being compatible with Python 3.12 `asyncio.wait_for`
        # https://github.com/python/cpython/pull/113850
        if isinstance(self.error, asyncio.CancelledError):
            return asyncio.CancelledError(f"[address={self.address}, pid={self.pid}]").with_traceback(self.traceback)

        return wrap_exception(
            self.error,
            (_AsCauseBase,),
            traceback=self.traceback,
            attr_dict=dict(address=self.address, pid=self.pid),
        )

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        item.serialized += (self.address, self.pid)
        item.subs = [self.error_type, self.error, self.traceback]
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        _MessageBase.deserial_members(self, serialized, subs)
        self.address = serialized[-2]
        self.pid = serialized[-1]
        self.error_type = subs[0]
        self.error = subs[1]
        self.traceback = subs[2]


cdef class CreateActorMessage(_MessageBase):
    message_type = MessageType.create_actor

    cdef:
        public type actor_cls
        public bytes actor_id
        public tuple args
        public dict kwargs
        public object allocate_strategy
        public object from_main

    def __init__(
        self,
        bytes message_id = None,
        type actor_cls = None,
        bytes actor_id = None,
        tuple args = None,
        dict kwargs = None,
        object allocate_strategy = None,
        object from_main: bool = False,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
    ):
        _MessageBase.__init__(
            self, message_id, protocol=protocol, message_trace=message_trace
        )
        self.actor_cls = actor_cls
        self.actor_id = actor_id
        self.args = args
        self.kwargs = kwargs
        self.allocate_strategy = allocate_strategy
        self.from_main = from_main

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        item.serialized += (
            self.actor_id, self.allocate_strategy, self.from_main
        )
        item.subs = [self.actor_cls, self.args, self.kwargs]
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        _MessageBase.deserial_members(self, serialized, subs)
        self.actor_id = serialized[-3]
        self.allocate_strategy = serialized[-2]
        self.from_main = serialized[-1]
        self.actor_cls = subs[0]
        self.args = subs[1]
        self.kwargs = subs[2]


cdef class DestroyActorMessage(_MessageBase):
    message_type = MessageType.destroy_actor

    cdef:
        public ActorRef actor_ref
        public object from_main

    def __init__(
        self,
        bytes message_id = None,
        ActorRef actor_ref = None,
        object from_main: bool = False,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
    ):
        _MessageBase.__init__(
            self, message_id, protocol=protocol, message_trace=message_trace
        )
        self.actor_ref = actor_ref
        self.from_main = from_main

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        item.serialized += (
            self.actor_ref.address, self.actor_ref.uid,
            self.actor_ref.proxy_addresses, self.from_main
        )
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        _MessageBase.deserial_members(self, serialized, subs)
        self.actor_ref = ActorRef(serialized[-4], serialized[-3], serialized[-2])
        self.from_main = serialized[-1]


cdef class HasActorMessage(_MessageBase):
    message_type = MessageType.has_actor

    cdef:
        public ActorRef actor_ref

    def __init__(
        self,
        bytes message_id = None,
        ActorRef actor_ref = None,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
    ):
        _MessageBase.__init__(
            self, message_id, protocol=protocol, message_trace=message_trace
        )
        self.actor_ref = actor_ref

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        item.serialized += (
            self.actor_ref.address, self.actor_ref.uid, self.actor_ref.proxy_addresses
        )
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        _MessageBase.deserial_members(self, serialized, subs)
        self.actor_ref = ActorRef(serialized[-3], serialized[-2], serialized[-1])


cdef class ActorRefMessage(_MessageBase):
    message_type = MessageType.actor_ref

    cdef:
        public ActorRef actor_ref

    def __init__(
        self,
        bytes message_id = None,
        ActorRef actor_ref = None,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
    ):
        _MessageBase.__init__(
            self, message_id, protocol=protocol, message_trace=message_trace
        )
        self.actor_ref = actor_ref

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        item.serialized += (
            self.actor_ref.address, self.actor_ref.uid, self.actor_ref.proxy_addresses
        )
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        _MessageBase.deserial_members(self, serialized, subs)
        self.actor_ref = ActorRef(serialized[-3], serialized[-2], serialized[-1])


cdef class SendMessage(_MessageBase):
    message_type = MessageType.send

    cdef:
        public ActorRef actor_ref
        public object content

    def __init__(
        self,
        bytes message_id = None,
        ActorRef actor_ref = None,
        object content = None,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
        object profiling_context = None,
    ):
        _MessageBase.__init__(
            self,
            message_id,
            protocol=protocol,
            message_trace=message_trace,
            profiling_context=profiling_context,
        )
        self.actor_ref = actor_ref
        self.content = content

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        item.serialized += (
            self.actor_ref.address, self.actor_ref.uid, self.actor_ref.proxy_addresses
        )
        item.subs = [self.content]
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        _MessageBase.deserial_members(self, serialized, subs)
        self.actor_ref = ActorRef(serialized[-3], serialized[-2], serialized[-1])
        self.content = subs[0]


cdef class TellMessage(SendMessage):
    message_type = MessageType.tell


cdef class CancelMessage(_MessageBase):
    message_type = MessageType.cancel

    cdef:
        public str address
        public bytes cancel_message_id

    def __init__(
        self,
        bytes message_id = None,
        str address = None,
        bytes cancel_message_id = None,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
    ):
        _MessageBase.__init__(
            self, message_id, protocol=protocol, message_trace=message_trace
        )
        self.address = address
        self.cancel_message_id = cancel_message_id

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        item.serialized += (
            self.address, self.cancel_message_id
        )
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        _MessageBase.deserial_members(self, serialized, subs)
        self.address = serialized[-2]
        self.cancel_message_id = serialized[-1]


cdef class CopyToBuffersMessage(_MessageBase):
    message_type = MessageType.copy_to_buffers

    cdef:
        public object content

    def __init__(
        self,
        bytes message_id = None,
        object content = None,
        int protocol = _DEFAULT_PROTOCOL,
        list message_trace = None,
    ):
        _MessageBase.__init__(
            self,
            message_id,
            protocol=protocol,
            message_trace=message_trace
        )
        self.content = content

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        item.subs = [self.content]
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        _MessageBase.deserial_members(self, serialized, subs)
        self.content = subs[0]


cdef class CopyToFileObjectsMessage(CopyToBuffersMessage):
    message_type = MessageType.copy_to_fileobjs


cdef class ForwardMessage(_MessageBase):
    message_type =  MessageType.forward

    cdef:
        public str address
        public _MessageBase raw_message

    def __init__(
        self,
        bytes message_id = None,
        str address = None,
        _MessageBase raw_message = None,
        int protocol = DEFAULT_PROTOCOL,
        list message_trace = None,
    ):
        _MessageBase.__init__(
            self,
            message_id,
            protocol=protocol,
            message_trace=message_trace
        )
        self.address = address
        self.raw_message = raw_message

    cdef _MessageSerialItem serial(self):
        cdef _MessageSerialItem item = _MessageBase.serial(self)
        cdef _MessageSerialItem raw_message_serialized = self.raw_message.serial()
        item.serialized += (self.address,)
        item.serialized += raw_message_serialized.serialized
        item.subs += raw_message_serialized.subs
        return item

    cdef deserial_members(self, tuple serialized, list subs):
        # 5 is magic number that means serialized for _MessageBase
        base_serialized = serialized[:5]
        _MessageBase.deserial_members(self, base_serialized, [])
        self.address = serialized[5]
        # process raw message
        tp = _message_type_to_message_cls[serialized[6]]
        cdef _MessageBase raw_message = <_MessageBase>(tp())
        raw_message.deserial_members(serialized[6:], subs)
        self.raw_message = raw_message


cdef dict _message_type_to_message_cls = {
    MessageType.control.value: ControlMessage,
    MessageType.result.value: ResultMessage,
    MessageType.error.value: ErrorMessage,
    MessageType.create_actor.value: CreateActorMessage,
    MessageType.destroy_actor.value: DestroyActorMessage,
    MessageType.has_actor.value: HasActorMessage,
    MessageType.actor_ref.value: ActorRefMessage,
    MessageType.send.value: SendMessage,
    MessageType.tell.value: TellMessage,
    MessageType.cancel.value: CancelMessage,
    MessageType.copy_to_buffers.value: CopyToBuffersMessage,
    MessageType.copy_to_fileobjs.value: CopyToFileObjectsMessage,
    MessageType.forward.value: ForwardMessage,
}


class DeserializeMessageFailed(RuntimeError):
    def __init__(self, message_id):
        self.message_id = message_id

    def __str__(self):
        return f"Deserialize {self.message_id} failed"


cdef class MessageSerializer(Serializer):
    serializer_id = 32105

    cpdef serial(self, object obj, dict context):
        cdef _MessageBase msg = <_MessageBase>obj
        cdef _MessageSerialItem ser_item

        assert msg.protocol == _DEFAULT_PROTOCOL, "only support protocol 0 for now"
        ser_item = msg.serial()
        return ser_item.serialized, ser_item.subs, False

    cpdef deserial(self, tuple serialized, dict context, list subs):
        cdef _MessageBase msg

        msg_type = serialized[0]
        msg = _message_type_to_message_cls[msg_type]()
        msg.deserial_members(serialized, subs)
        return msg

    cpdef on_deserial_error(
        self,
        tuple serialized,
        dict context,
        list subs_serialized,
        int error_index,
        object exc,
    ):
        message_id = serialized[1]  # pos of message_id field
        try:
            raise DeserializeMessageFailed(message_id) from exc
        except BaseException as new_ex:
            return new_ex


# register message serializer
MessageSerializer.register(_MessageBase)


cpdef bytes new_message_id():
    return new_random_id(32)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import concurrent.futures as futures
import logging
import os
import socket
import sys
from abc import ABCMeta
from asyncio import AbstractServer, StreamReader, StreamWriter
from functools import lru_cache
from hashlib import md5
from typing import Any, Callable, Coroutine, Dict, Type
from urllib.parse import urlparse

from ..._utils import to_binary
from ...constants import XOSCAR_CONNECT_TIMEOUT, XOSCAR_UNIX_SOCKET_DIR
from ...serialization import AioDeserializer, AioSerializer, deserialize
from ...utils import classproperty, implements, is_py_312, is_py_312_or_above, is_v6_ip
from .base import Channel, ChannelType, Client, Server
from .core import register_client, register_server
from .errors import ChannelClosed
from .utils import read_buffers, write_buffers

_is_windows: bool = sys.platform.startswith("win")


logger = logging.getLogger(__name__)


class SocketChannel(Channel):
    __slots__ = "reader", "writer", "_channel_type", "_send_lock", "_recv_lock"

    name = "socket"

    def __init__(
        self,
        reader: StreamReader,
        writer: StreamWriter,
        local_address: str | None = None,
        dest_address: str | None = None,
        compression: str | None = None,
        channel_type: int | None = None,
    ):
        super().__init__(
            local_address=local_address,
            dest_address=dest_address,
            compression=compression,
        )
        self.reader = reader
        self.writer = writer
        self._channel_type = channel_type

        self._send_lock = asyncio.Lock()
        self._recv_lock = asyncio.Lock()

    @property
    @implements(Channel.type)
    def type(self) -> int:
        return self._channel_type  # type: ignore

    @implements(Channel.send)
    async def send(self, message: Any):
        # get buffers
        compress = self.compression or 0
        serializer = AioSerializer(message, compress=compress)
        buffers = await serializer.run()

        try:
            # write buffers
            write_buffers(self.writer, buffers)
            async with self._send_lock:
                # add lock, or when parallel send,
                # assertion error may be raised
                await self.writer.drain()
        except RuntimeError as e:
            if self.writer.is_closing():
                raise ChannelClosed(
                    "Channel already closed, cannot write message"
                ) from e
            raise e

    @implements(Channel.recv)
    async def recv(self):
        deserializer = AioDeserializer(self.reader)
        async with self._recv_lock:
            header = await deserializer.get_header()
            buffers = await read_buffers(header, self.reader)
        return deserialize(header, buffers)

    @implements(Channel.close)
    async def close(self):
        self.writer.close()
        try:
            await self.writer.wait_closed()
        # TODO: May raise Runtime error: attach to another event loop
        except (ConnectionResetError, RuntimeError):  # pragma: no cover
            pass

    @property
    @implements(Channel.closed)
    def closed(self):
        return self.writer.is_closing()


class _BaseSocketServer(Server, metaclass=ABCMeta):
    __slots__ = "_aio_server", "_channels"

    _channels: set[Channel]

    def __init__(
        self,
        address: str,
        aio_server: AbstractServer,
        channel_handler: Callable[[Channel], Coroutine] | None = None,
    ):
        super().__init__(address, channel_handler)
        # asyncio.Server
        self._aio_server = aio_server
        self._channels = set()

    @implements(Server.start)
    async def start(self):
        await self._aio_server.start_serving()

    @implements(Server.join)
    async def join(self, timeout=None):
        if timeout is None:
            await self._aio_server.serve_forever()
        else:
            if is_py_312():
                # For python 3.12, there's a bug for `serve_forever`:
                # https://github.com/python/cpython/issues/123720,
                # which is unable to be cancelled.
                # Here is really a simulation of `wait_for`
                task = asyncio.create_task(self._aio_server.serve_forever())
                await asyncio.sleep(timeout)
                if task.done():
                    logger.warning(f"`serve_forever` should never be done.")
                else:
                    task.cancel()
            else:
                future = asyncio.create_task(self._aio_server.serve_forever())
                try:
                    await asyncio.wait_for(future, timeout=timeout)
                except (futures.TimeoutError, asyncio.TimeoutError, TimeoutError):
                    future.cancel()

    @implements(Server.on_connected)
    async def on_connected(self, *args, **kwargs):
        reader, writer = args
        local_address = kwargs.pop("local_address", None)
        dest_address = kwargs.pop("dest_address", None)
        if kwargs:  # pragma: no cover
            raise TypeError(
                f"{type(self).__name__} got unexpected "
                f'arguments: {",".join(kwargs)}'
            )
        channel = SocketChannel(
            reader,
            writer,
            local_address=local_address,
            dest_address=dest_address,
            channel_type=self.channel_type,
        )
        self._channels.add(channel)
        # handle over channel to some handlers
        try:
            await self.channel_handler(channel)
        finally:
            if not channel.closed:
                await channel.close()
            # Remove channel if channel exit
            self._channels.discard(channel)
            logger.debug("Channel exit: %s", channel.info)

    @implements(Server.stop)
    async def stop(self):
        self._aio_server.close()
        # Python 3.12+: # https://github.com/python/cpython/issues/104344
        # `wait_closed` leads to hang in Python 3.12 and 3.13
        if not is_py_312_or_above():
            await self._aio_server.wait_closed()
        # close all channels
        await asyncio.gather(
            *(channel.close() for channel in self._channels if not channel.closed)
        )
        self._channels.clear()

    @property
    @implements(Server.stopped)
    def stopped(self) -> bool:
        return not self._aio_server.is_serving()


@register_server
class SocketServer(_BaseSocketServer):
    __slots__ = "host", "port"

    scheme = None

    def __init__(
        self,
        host: str,
        port: int,
        aio_server: AbstractServer,
        channel_handler: Callable[[Channel], Coroutine] | None = None,
    ):
        address = f"{host}:{port}"
        super().__init__(address, aio_server, channel_handler=channel_handler)
        self.host = host
        self.port = port

    @classproperty
    @implements(Server.client_type)
    def client_type(self) -> Type["Client"]:
        return SocketClient

    @property
    @implements(Server.channel_type)
    def channel_type(self) -> int:
        return ChannelType.remote

    @classmethod
    def parse_config(cls, config: dict) -> dict:
        if config is None or not config:
            return dict()
        # we only need the following config
        keys = ["listen_elastic_ip"]
        parsed_config = {key: config[key] for key in keys if key in config}

        return parsed_config

    @staticmethod
    @implements(Server.create)
    async def create(config: Dict) -> "Server":
        config = config.copy()
        if "address" in config:
            address = config.pop("address")
            host, port = address.rsplit(":", 1)
            port = int(port)
        else:
            host = config.pop("host")
            port = int(config.pop("port"))
        _host = host
        if config.pop("listen_elastic_ip", False):
            # The Actor.address will be announce to client, and is not on our host,
            # cannot actually listen on it,
            # so we have to keep SocketServer.host untouched to make sure Actor.address not changed
            if is_v6_ip(host):
                _host = "::"
            else:
                _host = "0.0.0.0"

        handle_channel = config.pop("handle_channel")
        if "start_serving" not in config:
            config["start_serving"] = False

        async def handle_connection(reader: StreamReader, writer: StreamWriter):
            # create a channel when client connected
            return await server.on_connected(
                reader, writer, local_address=server.address
            )

        port = port if port != 0 else None
        aio_server = await asyncio.start_server(
            handle_connection, host=_host, port=port, **config
        )

        # get port of the socket if not specified
        if not port:
            port = aio_server.sockets[0].getsockname()[1]

        if _is_windows:
            for sock in aio_server.sockets:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)

        server = SocketServer(host, port, aio_server, channel_handler=handle_channel)
        return server


@register_client
class SocketClient(Client):
    __slots__ = ()

    scheme = SocketServer.scheme

    @staticmethod
    @implements(Client.connect)
    async def connect(
        dest_address: str, local_address: str | None = None, **kwargs
    ) -> "Client":
        host, port_str = dest_address.rsplit(":", 1)
        port = int(port_str)
        config = kwargs.get("config", {})
        connect_timeout = config.get("connect_timeout", XOSCAR_CONNECT_TIMEOUT)
        fut = asyncio.open_connection(host=host, port=port)
        try:
            reader, writer = await asyncio.wait_for(fut, timeout=connect_timeout)
        except asyncio.TimeoutError:
            raise ConnectionError("connect timeout")
        channel = SocketChannel(
            reader,
            writer,
            local_address=local_address,
            dest_address=dest_address,
            channel_type=ChannelType.remote,
        )
        return SocketClient(local_address, dest_address, channel)


def _get_or_create_default_unix_socket_dir():
    os.makedirs(XOSCAR_UNIX_SOCKET_DIR, exist_ok=True)
    return XOSCAR_UNIX_SOCKET_DIR


@lru_cache(100)
def _gen_unix_socket_default_path(process_index):
    return (
        f"{_get_or_create_default_unix_socket_dir()}/"
        f"{md5(to_binary(str(process_index))).hexdigest()}"
    )  # nosec


@register_server
class UnixSocketServer(_BaseSocketServer):
    __slots__ = "process_index", "path"

    scheme = "unixsocket"

    def __init__(
        self,
        process_index: int,
        aio_server: AbstractServer,
        path: str,
        channel_handler: Callable[[Channel], Coroutine] | None = None,
    ):
        address = f"{self.scheme}:///{process_index}"
        super().__init__(address, aio_server, channel_handler=channel_handler)
        self.process_index = process_index
        self.path = path

    @classproperty
    @implements(Server.client_type)
    def client_type(self) -> Type["Client"]:
        return UnixSocketClient

    @property
    @implements(Server.channel_type)
    def channel_type(self) -> int:
        return ChannelType.ipc

    @staticmethod
    @implements(Server.create)
    async def create(config: Dict) -> "Server":
        config = config.copy()
        if "address" in config:
            process_index = int(urlparse(config.pop("address")).path.lstrip("/"))
        else:
            process_index = config.pop("process_index")
        handle_channel = config.pop("handle_channel")
        path = config.pop("path", _gen_unix_socket_default_path(process_index))

        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname, exist_ok=True)

        if "start_serving" not in config:
            config["start_serving"] = False

        async def handle_connection(reader, writer):
            # create a channel when client connected
            return await server.on_connected(
                reader, writer, local_address=server.address
            )

        aio_server = await asyncio.start_unix_server(
            handle_connection, path=path, **config
        )

        for sock in aio_server.sockets:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)

        server = UnixSocketServer(
            process_index, aio_server, path, channel_handler=handle_channel
        )
        return server

    @implements(Server.stop)
    async def stop(self):
        await super().stop()
        try:
            os.remove(self.path)
        except OSError:  # pragma: no cover
            pass


@register_client
class UnixSocketClient(Client):
    __slots__ = ()

    scheme = UnixSocketServer.scheme

    @staticmethod
    @lru_cache(100)
    def _get_process_index(addr):
        return int(urlparse(addr).path.lstrip("/"))

    @staticmethod
    @implements(Client.connect)
    async def connect(
        dest_address: str, local_address: str | None = None, **kwargs
    ) -> "Client":
        process_index = UnixSocketClient._get_process_index(dest_address)
        path = kwargs.pop("path", _gen_unix_socket_default_path(process_index))
        try:
            (reader, writer) = await asyncio.open_unix_connection(path, **kwargs)
        except FileNotFoundError:
            raise ConnectionRefusedError(
                "Cannot connect unix socket due to file not exists"
            )
        channel = SocketChannel(
            reader,
            writer,
            local_address=local_address,
            dest_address=dest_address,
            channel_type=ChannelType.ipc,
        )
        return UnixSocketClient(local_address, dest_address, channel)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import (
    AioBase,
    delegate_to_executor,
    proxy_method_directly,
    proxy_property_directly,
)


@delegate_to_executor(
    "close",
    "flush",
    "isatty",
    "read",
    "read1",
    "readinto",
    "readline",
    "readlines",
    "seek",
    "seekable",
    "tell",
    "truncate",
    "writable",
    "write",
    "writelines",
)
@proxy_method_directly("fileno", "readable")
@proxy_property_directly("closed", "name", "mode")
class AioFileObject(AioBase):
    def __aiter__(self):
        return self

    async def __anext__(self):
        """Simulate normal file iteration."""
        line = await self.readline()
        if line:
            return line
        else:
            raise StopAsyncIteration

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
        self._file = None



# Copyright 2022-2025 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List

import numpy as np

from ..utils import lazy_import
from .core import Serializer, buffered

mx = lazy_import("mlx.core")


dtype_map = {
    "b": np.int8,
    "B": np.uint8,
    "h": np.int16,
    "H": np.uint16,
    "i": np.int32,
    "I": np.uint32,
    "q": np.int64,
    "Q": np.uint64,
    "e": np.float16,
    "f": np.float32,
    "d": np.float64,
}


class MLXSerislizer(Serializer):
    @buffered
    def serial(self, obj: "mx.array", context: dict):  # type: ignore
        ravel_obj = obj.reshape(-1).view(mx.uint8)
        mv = memoryview(ravel_obj)
        header = dict(
            shape=obj.shape, format=mv.format, dtype=str(obj.dtype).rsplit(".", 1)[-1]
        )
        if not mv.c_contiguous:
            # NOTE: we only consider c contiguous here,
            # MLX has no way to create f contiguous arrays.
            mv = memoryview(bytes(mv))
        return (header,), [mv], True

    def deserial(self, serialized: tuple, context: dict, subs: List[Any]):
        header = serialized[0]
        shape, format, dtype = header["shape"], header["format"], header["dtype"]
        mv = memoryview(subs[0])
        if mv.format != format:
            dtype = dtype_map.get(format, np.uint8)
            np_arr = np.frombuffer(mv, dtype=dtype).reshape(shape)  # parse
            mv = memoryview(np_arr)  # recreate memoryview
        ravel_array = mx.array(mv)
        return ravel_array.view(getattr(mx, dtype)).reshape(shape)


if mx is not None:
    MLXSerislizer.register(mx.array)



# Copyright 2022-2025 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import ast
import importlib
import json
import operator
from abc import ABC, abstractmethod
from pathlib import Path

from packaging.markers import Marker, default_environment
from packaging.requirements import InvalidRequirement, Requirement
from packaging.version import Version

from .platform import (
    check_cuda_available,
    check_npu_available,
    get_cuda_arch,
    get_cuda_version,
)
from .utils import is_vcs_url


def resolve_system_requirement(req_part: str) -> str:
    req_part = req_part.strip()
    if not (req_part.startswith("#system_") and req_part.endswith("#")):
        return req_part

    real_pkg = req_part[len("#system_") : -1]
    try:
        version = importlib.metadata.version(real_pkg)
        version = version.split("+")[0]
    except importlib.metadata.PackageNotFoundError:
        raise RuntimeError(
            f"System package '{real_pkg}' not found. Cannot resolve '{req_part}'."
        )
    return f"{real_pkg}=={version}"


class VirtualEnvManager(ABC):
    @classmethod
    @abstractmethod
    def is_available(cls):
        pass

    def __init__(self, env_path: Path):
        self.env_path = env_path.resolve()

    @abstractmethod
    def exists_env(self) -> bool:
        pass

    @abstractmethod
    def create_env(
        self, python_path: Path | None = None, exists: str = "ignore"
    ) -> None:
        pass

    @abstractmethod
    def install_packages(self, packages: list[str], **kwargs):
        pass

    @staticmethod
    def process_packages(packages: list[str], **variables) -> list[str]:
        """
        Process a list of package names, replacing placeholders like #system_<package>#
        with the installed version of the corresponding package from the system environment.

        Example:
            "#system_torch#" -> "torch==2.1.0" (if torch 2.1.0 is installed)

        Args:
            packages (list[str]): A list of package names, which may include placeholders.
            **variables: Dynamic variables for marker substitution, e.g., engine='vllm'.

        Returns:
            list[str]: A new list with resolved package names and versions.

        Raises:
            RuntimeError: If a specified system package is not found in the environment.
        """
        processed = []

        for pkg in packages:
            processed.append(resolve_system_requirement(pkg))

        # apply extended syntax including:
        # - has_cuda: whether CUDA is available (bool)
        # - cuda_version: CUDA version string, e.g. "12.1" (str)
        # - cuda_arch: CUDA architecture string, e.g. "sm_80" (str)
        # - has_npu: whether an NPU is available (bool)
        # - #var#: dynamic variable substitution, e.g., #engine# == "vllm"
        processed = filter_requirements(processed, **variables)

        return processed

    @abstractmethod
    def cancel_install(self):
        pass

    @abstractmethod
    def get_python_path(self) -> str | None:
        pass

    @abstractmethod
    def get_lib_path(self) -> str:
        pass

    @abstractmethod
    def remove_env(self):
        pass


def substitute_variables(marker_str: str, variables: dict) -> str:
    """
    Replace #var# placeholders in marker string with actual values.

    Example:
        substitute_variables('#engine# == "vllm"', {'engine': 'vllm'})
        -> '"vllm" == "vllm"'

    Args:
        marker_str: Marker expression, e.g., '#engine# == "vllm"'
        variables: Variable dict, e.g., {'engine': 'vllm', 'count': 10}

    Returns:
        Marker string with placeholders replaced
    """
    result = marker_str

    for var_name, var_value in variables.items():
        placeholder = f"#{var_name}#"
        if placeholder not in result:
            continue

        # Format value based on type
        if var_value is None:
            formatted = "None"
        elif isinstance(var_value, bool):
            # Boolean: use Python literal (True/False) without quotes
            formatted = str(var_value)  # True -> "True", False -> "False"
        elif isinstance(var_value, (int, float)):
            formatted = str(var_value)
        else:
            # String: escape and add quotes
            formatted = json.dumps(var_value)

        result = result.replace(placeholder, formatted)

    return result


def get_env() -> dict[str, str | bool | int | float | None]:
    env = default_environment().copy()
    # Your custom env vars here, e.g.:
    env.update(
        {
            "has_cuda": check_cuda_available(),
            "cuda_version": get_cuda_version(),
            "cuda_arch": get_cuda_arch(),
            "has_npu": check_npu_available(),
        }
    )
    return env


STANDARD_ENV_VARS = set(default_environment().keys())


def is_custom_marker(marker_str: str) -> bool:
    try:
        marker = Marker(marker_str)
    except Exception:
        return True

    def traverse_markers(node):
        if isinstance(node, tuple):
            env_var = node[0]
            if env_var not in STANDARD_ENV_VARS:
                return True
            return False
        elif isinstance(node, list):
            return any(traverse_markers(child) for child in node)
        return False

    return traverse_markers(marker._markers)


def eval_custom_marker(marker_str: str, env: dict) -> bool:
    ops = {
        ast.Eq: operator.eq,
        ast.NotEq: operator.ne,
        ast.Lt: operator.lt,
        ast.LtE: operator.le,
        ast.Gt: operator.gt,
        ast.GtE: operator.ge,
        ast.And: lambda a, b: a and b,
        ast.Or: lambda a, b: a or b,
        ast.Not: operator.not_,
    }

    def normalize_value(val):
        # Normalize for boolean
        if isinstance(val, str):
            if val.lower() == "true":
                return True
            if val.lower() == "false":
                return False

        # Normalize for version-like fields
        if isinstance(val, str):
            if val.count(".") >= 1 and all(
                part.isdigit() for part in val.split(".") if part
            ):
                return Version(val)

        return val

    def maybe_parse_cuda_arch(val):
        if isinstance(val, str) and val.startswith("sm_"):
            try:
                return int(val[3:])
            except ValueError:
                return val
        return val

    def _eval(node):
        if isinstance(node, ast.BoolOp):
            left = _eval(node.values[0])
            for right_node in node.values[1:]:
                right = _eval(right_node)
                op = ops[type(node.op)]
                left = op(left, right)
            return left

        elif isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.Not):
            return not _eval(node.operand)

        elif isinstance(node, ast.Compare):
            left = _eval(node.left)
            left = maybe_parse_cuda_arch(normalize_value(left))

            for op_node, right_expr in zip(node.ops, node.comparators):
                right = _eval(right_expr)
                right = maybe_parse_cuda_arch(normalize_value(right))

                op_func = ops[type(op_node)]
                if not op_func(left, right):
                    return False
                left = right  # for chained comparisons

            return True

        elif isinstance(node, ast.Name):
            return normalize_value(env.get(node.id))

        elif isinstance(node, ast.Constant):
            return node.value

        elif isinstance(node, ast.Str):  # Python <3.8
            return node.s

        else:
            raise ValueError(f"Unsupported expression: {ast.dump(node)}")

    tree = ast.parse(marker_str, mode="eval")
    return _eval(tree.body)


def filter_requirements(requirements: list[str], **variables) -> list[str]:
    """
    Filter requirements by evaluating markers in given env.
    If env is None, use get_env().

    Args:
        requirements: List of requirement strings
        **variables: Dynamic variables for #var# substitution, e.g., engine='vllm'
    """
    env = get_env()
    for var_name, var_value in variables.items():
        if isinstance(var_value, (str, int, float, bool)) or var_value is None:
            env[var_name] = var_value
    result = []
    for req_str in requirements:
        if is_vcs_url(req_str):
            result.append(req_str)
        elif ";" in req_str:
            req_part, marker_part = req_str.split(";", 1)
            marker_part = marker_part.strip()
            req_part = resolve_system_requirement(req_part.strip())

            # Substitute #var# placeholders with actual values
            marker_part = substitute_variables(marker_part, variables)

            try:
                req = Requirement(f"{req_part}; {marker_part}")
                try:
                    marker_ok = req.marker is None or req.marker.evaluate(env)
                except Exception:
                    marker_ok = None
                if marker_ok:
                    result.append(f"{req.name}{req.specifier}")
                    continue
                if marker_ok is None and is_custom_marker(marker_part):
                    if eval_custom_marker(marker_part, env):
                        req = Requirement(req_part)
                        result.append(str(req))
                    continue
            except InvalidRequirement:
                if is_custom_marker(marker_part):
                    if eval_custom_marker(marker_part, env):
                        req = Requirement(req_part)
                        result.append(str(req))
                else:
                    raise
        else:
            req = Requirement(req_str.strip())
            result.append(str(req))

    return result



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Coroutine, Type

from ...utils import classproperty, implements


@dataclass
class ChannelType:
    local = 0  # for local communication
    ipc = 1  # inproc
    remote = 2  # remote


class Channel(ABC):
    """
    Channel is used to do data exchange between server and client.
    """

    __slots__ = "local_address", "dest_address", "compression"

    name: str | None = None

    def __init__(
        self,
        local_address: str | None = None,
        dest_address: str | None = None,
        compression: str | None = None,
    ):
        self.local_address = local_address
        self.dest_address = dest_address
        self.compression = compression

    @abstractmethod
    async def send(self, message: Any):
        """
        Send data to dest. There should be only one send for one recv, otherwise recv messages
        may overlap.

        Parameters
        ----------
        message:
            data that sent to dest.
        """

    @abstractmethod
    async def recv(self):
        """
        Receive data that sent from dest.
        """

    @abstractmethod
    async def close(self):
        """
        Close channel.
        """

    @property
    @abstractmethod
    def closed(self) -> bool:
        """
        This channel is closed or not.

        Returns
        -------
        closed:
            If the channel is closed.
        """

    @property
    @abstractmethod
    def type(self) -> int:
        """
        Channel is used for, can be dummy, ipc or remote.

        Returns
        -------
        channel_type: int
            type that can be dummy, ipc or remote.
        """

    @property
    def info(self) -> dict:
        return {
            "name": self.name,
            "compression": self.compression,
            "type": self.type,
            "local_address": self.local_address,
            "dest_address": self.dest_address,
        }


class Server(ABC):
    __slots__ = "address", "channel_handler"

    scheme: str | None = None

    def __init__(
        self,
        address: str,
        channel_handler: Callable[[Channel], Coroutine] | None = None,
    ):
        self.address = address
        self.channel_handler = channel_handler

    @classproperty
    @abstractmethod
    def client_type(self) -> Type["Client"]:
        """
        Return the corresponding client type.

        Returns
        -------
        client_type: type
            client type.
        """

    @property
    @abstractmethod
    def channel_type(self) -> int:
        """
        Channel type, can be dummy, ipc or remote.

        Returns
        -------
        channel_type: int
            type that can be dummy, ipc or remote.
        """

    @staticmethod
    @abstractmethod
    async def create(config: dict) -> "Server":
        """
        Create a server instance according to configuration.

        Parameters
        ----------
        config: dict
            configuration about creating a channel.

        Returns
        -------
        server: Server
            a server that waiting for connections from clients.
        """

    @abstractmethod
    async def start(self):
        """
        Used for listening to port or similar stuff.
        """

    @abstractmethod
    async def join(self, timeout=None):
        """
        Wait forever until timeout.
        """

    @abstractmethod
    async def on_connected(self, *args, **kwargs):
        """
        Return a channel when new client connected.

        Returns
        -------
        channel: Channel
            channel for communication
        """

    @abstractmethod
    async def stop(self):
        """
        Stop the server.
        """

    @property
    @abstractmethod
    def stopped(self) -> bool:
        """
        If this server is stopped or not.

        Returns
        -------
        if_stopped: bool
           This server is stopped or not.
        """

    @property
    def info(self) -> dict:
        return {
            "name": self.scheme,
            "address": self.address,
            "channel_type": self.channel_type,
        }

    @classmethod
    def parse_config(cls, config: dict) -> dict:
        # skip parsing config by default
        return dict()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()


class Client(ABC):
    __slots__ = "local_address", "dest_address", "channel"

    scheme: str | None = None

    def __init__(
        self, local_address: str | None, dest_address: str | None, channel: Channel
    ):
        self.local_address = local_address
        self.dest_address = dest_address
        self.channel = channel

    @property
    def channel_type(self) -> int:
        """
        Channel type, can be dummy, ipc or remote.

        Returns
        -------
        channel_type: int
            type that can be dummy, ipc or remote.
        """
        return self.channel.type

    @staticmethod
    @abstractmethod
    async def connect(
        dest_address: str, local_address: str | None = None, **kwargs
    ) -> "Client":
        """
        Create a client that is able to connect to some server.

        Parameters
        ----------
        dest_address: str
            Destination server address that to connect to.
        local_address: str
            local address.

        Returns
        -------
        client: Client
            Client that holds a channel to communicate.
        """

    @classmethod
    def parse_config(cls, config: dict) -> dict:
        # skip parsing config by default
        return dict()

    @implements(Channel.send)
    async def send(self, message):
        return await self.channel.send(message)

    @implements(Channel.recv)
    async def recv(self):
        return await self.channel.recv()

    async def close(self):
        """
        Close connection.
        """
        await self.channel.close()

    @property
    def closed(self) -> bool:
        """
        This client is closed or not.

        Returns
        -------
        closed: bool
            If the client is closed.
        """
        return self.channel.closed

    @property
    def info(self) -> dict:
        return {
            "local_address": self.local_address,
            "dest_address": self.dest_address,
            "channel_name": self.channel.name,
            "channel_type": self.channel_type,
        }

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import atexit
import copy
import logging
import threading
import weakref
from typing import Type, Union

from .._utils import Timer
from ..errors import ServerClosed
from ..profiling import get_profiling_data
from .communication import ChannelType, Client, UCXClient
from .message import (
    DeserializeMessageFailed,
    ErrorMessage,
    ForwardMessage,
    MessageType,
    ResultMessage,
    _MessageBase,
)
from .router import Router

ResultMessageType = Union[ResultMessage, ErrorMessage]
logger = logging.getLogger(__name__)


class ActorCallerThreadLocal:
    __slots__ = ("_client_to_message_futures", "_clients", "_profiling_data")

    _client_to_message_futures: dict[Client, dict[bytes, asyncio.Future]]
    _clients: dict[Client, asyncio.Task]

    def __init__(self):
        self._client_to_message_futures = dict()
        self._clients = dict()
        self._profiling_data = get_profiling_data()

    def _listen_client(self, client: Client):
        if client not in self._clients:
            self._clients[client] = asyncio.create_task(self._listen(client))
            self._client_to_message_futures[client] = dict()
            client_count = len(self._clients)
            if client_count >= 100:  # pragma: no cover
                if (client_count - 100) % 10 == 0:  # pragma: no cover
                    logger.warning(
                        "Actor caller has created too many clients (%s >= 100), "
                        "the global router may not be set.",
                        client_count,
                    )

    async def get_client_via_type(
        self, router: Router, dest_address: str, client_type: Type[Client]
    ) -> Client:
        client = await router.get_client_via_type(
            dest_address, client_type, from_who=self
        )
        self._listen_client(client)
        return client

    async def get_client(
        self,
        router: Router,
        dest_address: str,
        proxy_addresses: list[str] | None = None,
    ) -> Client:
        client = await router.get_client(
            dest_address, from_who=self, proxy_addresses=proxy_addresses
        )
        self._listen_client(client)
        return client

    async def _listen(self, client: Client):
        try:
            while not client.closed:
                try:
                    try:
                        message: _MessageBase = await client.recv()
                    except (EOFError, ConnectionError, BrokenPipeError) as e:
                        # AssertionError is from get_header
                        # remote server closed, close client and raise ServerClosed
                        logger.debug(f"{client.dest_address} close due to {e}")
                        try:
                            await client.close()
                        except (ConnectionError, BrokenPipeError):
                            # close failed, ignore it
                            pass
                        raise ServerClosed(
                            f"Remote server {client.dest_address} closed: {e}"
                        ) from None
                    future = self._client_to_message_futures[client].pop(
                        message.message_id
                    )
                    if not future.done():
                        future.set_result(message)
                except DeserializeMessageFailed as e:
                    message_id = e.message_id
                    future = self._client_to_message_futures[client].pop(message_id)
                    future.set_exception(e.__cause__)  # type: ignore
                except Exception as e:  # noqa: E722  # pylint: disable=bare-except
                    message_futures = self._client_to_message_futures[client]
                    self._client_to_message_futures[client] = dict()
                    for future in message_futures.values():
                        future.set_exception(copy.copy(e))
                finally:
                    # message may have Ray ObjectRef, delete it early in case next loop doesn't run
                    # as soon as expected.
                    try:
                        del message
                    except NameError:
                        pass
                    try:
                        del future
                    except NameError:
                        pass
                    await asyncio.sleep(0)

            message_futures = self._client_to_message_futures[client]
            self._client_to_message_futures[client] = dict()
            error = ServerClosed(f"Remote server {client.dest_address} closed")
            for future in message_futures.values():
                future.set_exception(copy.copy(error))
        finally:
            try:
                await client.close()
            except:  # noqa: E722  # nosec  # pylint: disable=bare-except
                # ignore all error if fail to close at last
                pass

    async def call_with_client(
        self, client: Client, message: _MessageBase, wait: bool = True
    ) -> ResultMessage | ErrorMessage | asyncio.Future:
        loop = asyncio.get_running_loop()
        wait_response = loop.create_future()
        self._client_to_message_futures[client][message.message_id] = wait_response

        with Timer() as timer:
            try:
                await client.send(message)
            except ConnectionError:
                try:
                    await client.close()
                except ConnectionError:
                    # close failed, ignore it
                    pass
                raise ServerClosed(f"Remote server {client.dest_address} closed")

            if not wait:
                r = wait_response
            else:
                r = await wait_response

        self._profiling_data.collect_actor_call(message, timer.duration)
        return r

    async def call_send_buffers(
        self,
        client: UCXClient,
        local_buffers: list,
        meta_message: _MessageBase,
        wait: bool = True,
    ) -> ResultMessage | ErrorMessage | asyncio.Future:
        loop = asyncio.get_running_loop()
        wait_response = loop.create_future()
        self._client_to_message_futures[client][meta_message.message_id] = wait_response

        with Timer() as timer:
            try:
                await client.send_buffers(local_buffers, meta_message)
            except ConnectionError:  # pragma: no cover
                try:
                    await client.close()
                except ConnectionError:
                    # close failed, ignore it
                    pass
                raise ServerClosed(f"Remote server {client.dest_address} closed")

            if not wait:  # pragma: no cover
                r = wait_response
            else:
                r = await wait_response

        self._profiling_data.collect_actor_call(meta_message, timer.duration)
        return r

    async def call(
        self,
        router: Router,
        dest_address: str,
        message: _MessageBase,
        wait: bool = True,
        proxy_addresses: list[str] | None = None,
    ) -> ResultMessage | ErrorMessage | asyncio.Future:
        client = await self.get_client(
            router, dest_address, proxy_addresses=proxy_addresses
        )
        if (
            client.channel_type == ChannelType.remote
            and client.dest_address != dest_address
            and message.message_type != MessageType.control
        ):
            # wrap message with forward message
            message = ForwardMessage(
                message_id=message.message_id, address=dest_address, raw_message=message
            )
        return await self.call_with_client(client, message, wait)

    async def stop(self):
        try:
            await asyncio.gather(*[client.close() for client in self._clients])
        except (ConnectionError, ServerClosed):
            pass
        try:
            self.cancel_tasks()
        except:
            pass

    def cancel_tasks(self):
        # cancel listening for all clients
        _ = [task.cancel() for task in self._clients.values()]


def _cancel_all_tasks(loop):
    to_cancel = asyncio.all_tasks(loop)
    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    # In Python 3.13+, we need to use a different approach to avoid deadlocks
    # when shutting down event loops in threads
    if hasattr(asyncio, "run"):
        # For Python 3.13+, use a more robust approach
        async def _gather_cancelled():
            await asyncio.gather(*to_cancel, return_exceptions=True)

        try:
            # Try to run the gather in the current loop context
            if loop.is_running():
                # If loop is running, schedule the gather
                asyncio.run_coroutine_threadsafe(_gather_cancelled(), loop)
            else:
                # If loop is not running, we can run it directly
                loop.run_until_complete(_gather_cancelled())
        except RuntimeError:
            # If we can't run the gather, just log and continue
            logger.debug("Could not gather cancelled tasks during shutdown")
    else:
        # For older Python versions, use the original approach
        loop.run_until_complete(asyncio.gather(*to_cancel, return_exceptions=True))

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during asyncio.run() shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )


def _safe_run_forever(loop):
    try:
        loop.run_forever()
    finally:
        try:
            _cancel_all_tasks(loop)
        except Exception as e:
            logger.debug("Error during task cancellation: %s", e)
        finally:
            try:
                loop.stop()
            except Exception as e:
                logger.debug("Error stopping loop: %s", e)


class ActorCaller:
    __slots__ = "_thread_local"

    class _RefHolder:
        pass

    _close_loop = None
    _close_thread = None
    _initialized = False

    @classmethod
    def _ensure_initialized(cls):
        if not cls._initialized:
            cls._close_loop = asyncio.new_event_loop()
            cls._close_thread = threading.Thread(
                target=_safe_run_forever, args=(cls._close_loop,), daemon=True
            )
            cls._close_thread.start()
            atexit.register(cls._cleanup)
            cls._initialized = True

    @classmethod
    def _cleanup(cls):
        if cls._close_loop and cls._close_loop.is_running():
            try:
                cls._close_loop.call_soon_threadsafe(cls._close_loop.stop)
                # Give the loop a moment to stop
                if cls._close_thread:
                    cls._close_thread.join(timeout=0.5)  # Shorter timeout for tests
            except Exception as e:
                logger.debug("Error during cleanup: %s", e)

    def __init__(self):
        self._thread_local = threading.local()

    def __getattr__(self, item):
        try:
            actor_caller = self._thread_local.actor_caller
        except AttributeError:
            thread_info = str(threading.current_thread())
            logger.debug("Creating a new actor caller for thread: %s", thread_info)
            actor_caller = self._thread_local.actor_caller = ActorCallerThreadLocal()
            ref = self._thread_local.ref = ActorCaller._RefHolder()
            # If the thread exit, we clean the related actor callers and channels.

            def _cleanup():
                self._ensure_initialized()
                # Use the background thread for cleanup
                asyncio.run_coroutine_threadsafe(actor_caller.stop(), self._close_loop)
                logger.debug(
                    "Clean up the actor caller due to thread exit: %s", thread_info
                )

            weakref.finalize(ref, _cleanup)

        return getattr(actor_caller, item)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any


class ActorPoolConfig:
    __slots__ = ("_conf",)

    def __init__(self, conf: dict | None = None):
        if conf is None:
            conf = dict()
        self._conf = conf
        if "pools" not in self._conf:
            self._conf["pools"] = dict()
        if "mapping" not in self._conf:
            self._conf["mapping"] = dict()
        if "metrics" not in self._conf:
            self._conf["metrics"] = dict()
        if "comm" not in self._conf:
            self._conf["comm"] = dict()
        if "proxy" not in self._conf:
            self._conf["proxy"] = dict()

    @property
    def n_pool(self):
        return len(self._conf["pools"])

    def add_pool_conf(
        self,
        process_index: int,
        label: str | None,
        internal_address: str | None,
        external_address: str | list[str],
        env: dict | None = None,
        modules: list[str] | None = None,
        suspend_sigint: bool | None = False,
        use_uvloop: bool | None = False,
        logging_conf: dict | None = None,
        kwargs: dict | None = None,
    ):
        pools: dict = self._conf["pools"]
        if not isinstance(external_address, list):
            external_address = [external_address]
        pools[process_index] = {
            "label": label,
            "internal_address": internal_address,
            "external_address": external_address,
            "env": env,
            "modules": modules,
            "suspend_sigint": suspend_sigint,
            "use_uvloop": use_uvloop,
            "logging_conf": logging_conf,
            "kwargs": kwargs or {},
        }

        mapping: dict = self._conf["mapping"]
        for addr in external_address:
            mapping[addr] = internal_address

    def remove_pool_config(self, process_index: int):
        addr = self.get_external_address(process_index)
        del self._conf["pools"][process_index]
        del self._conf["mapping"][addr]

    def get_pool_config(self, process_index: int):
        return self._conf["pools"][process_index]

    def get_external_address(self, process_index: int) -> str:
        return self._conf["pools"][process_index]["external_address"][0]

    def get_process_indexes(self):
        return list(self._conf["pools"])

    def get_process_index(self, external_address: str):
        for process_index, conf in self._conf["pools"].items():
            if external_address in conf["external_address"]:
                return process_index
        raise ValueError(
            f"Cannot get process_index for {external_address}"
        )  # pragma: no cover

    def reset_pool_external_address(
        self,
        process_index: int,
        external_address: str | list[str],
    ):
        if not isinstance(external_address, list):
            external_address = [external_address]
        cur_pool_config = self._conf["pools"][process_index]
        internal_address = cur_pool_config["internal_address"]

        mapping: dict = self._conf["mapping"]
        for addr in cur_pool_config["external_address"]:
            if internal_address == addr:
                # internal address may be the same as external address in Windows
                internal_address = external_address[0]
            mapping.pop(addr, None)

        cur_pool_config["external_address"] = external_address
        for addr in external_address:
            mapping[addr] = internal_address

    def get_external_addresses(self, label=None) -> list[str]:
        result = []
        for c in self._conf["pools"].values():
            if label is not None:
                if label == c["label"]:
                    result.append(c["external_address"][0])
            else:
                result.append(c["external_address"][0])
        return result

    @property
    def external_to_internal_address_map(self) -> dict[str, str]:
        return self._conf["mapping"]

    def as_dict(self):
        return self._conf

    def add_metric_configs(self, metrics: dict[str, Any]):
        if metrics:
            self._conf["metrics"].update(metrics)

    def get_metric_configs(self):
        return self._conf["metrics"]

    def add_comm_config(self, comm_config: dict[str, Any] | None):
        if comm_config:
            self._conf["comm"].update(comm_config)

    def get_comm_config(self) -> dict:
        return self._conf["comm"]

    def get_proxy_config(self) -> dict:
        return self._conf["proxy"]

    def add_proxy_config(self, proxy_config: dict[str, str] | None):
        if proxy_config:
            self._conf["proxy"].update(proxy_config)

    def remove_proxy(self, from_addr: str):
        del self._conf["proxy"][from_addr]
